"""
PDB structure handling based on MMTF format

Authors:
  Thomas A. Hopf
"""

from collections import OrderedDict, Iterable
from os import path
from urllib.error import HTTPError

from mmtf import fetch, parse
import numpy as np
import pandas as pd

from evcouplings.utils.config import InvalidParameterError
from evcouplings.utils.constants import AA3_to_AA1
from evcouplings.utils.helpers import DefaultOrderedDict
from evcouplings.utils.system import (
    valid_file, ResourceError, tempdir
)

# Mapping from MMTF secondary structure codes to DSSP symbols
MMTF_DSSP_CODE_MAP = {
    0: "I",  # pi helix
    1: "S",  # bend
    2: "H",  # alpha helix
    3: "E",  # extended
    4: "G",  # 3-10 helix
    5: "B",  # bridge
    6: "T",  # turn
    7: "C",  # coil
    -1: "",  # undefined
}

# Reduction to 3-state secondary structure annotation
DSSP_3_STATE_MAP = {
    "H": "H",
    "G": "H",
    "I": "H",
    "E": "E",
    "B": "E",
    "C": "C",
    "T": "C",
    "S": "C",
}

# format string for PDB ATOM records
PDB_FORMAT = (
    "{atom:<6s}{atom_id:>5} "
    "{atom_name:4s}{alt_loc_ind:1s}{residue_name:<3s} "
    "{chain_id:1s}{residue_id:>4}{ins_code:1}   "
    "{x_coord:>8.3f}{y_coord:>8.3f}{z_coord:>8.3f}"
    "{occupancy:>6.2f}{temp_factor:>6.2f}          "
    "{element_symbol:>2}{charge:>2}"
)


class Chain:
    """
    Container for PDB chain residue and coordinate information
    """
    def __init__(self, residues, coords):
        """
        Create new PDB chain, indexed by residue coordinate
        indeces

        Parameters
        ----------
        residues : pandas.DataFrame
            List of residues (as computed by PDB.get_chain())
        coords : pandas.DataFrame
            List of atom coordinates (as computed by
            PDB.get_chain())
        """
        self.residues = residues
        self.coords = coords

    def _update_ids(self, ids):
        """
        Update residue identifiers, and remove any
        residue that does not have new id. Also
        removes corresponding atom coordinates.

        Parameters
        ----------
        ids : pandas.Series (or list-like)
            New identifiers to be assigned to
            residue table. Has to be of same length
            in same order as residue table.

        Returns
        -------
        Chain
            Chain with new residue identifiers
        """
        residues = self.residues.copy()
        residues.loc[:, "id"] = ids.copy()
        residues = residues.dropna(subset=["id"])

        # drop coordinates of residues that were not kept
        coords = self.coords.loc[
            self.coords.residue_index.isin(residues.index)
        ]

        # reset atom index to consecutive numbers from 0
        coords = coords.reset_index(drop=True)

        return Chain(residues, coords)

    def to_seqres(self):
        """
        Return copy of chain with main index set to
        SEQRES numbering. Residues that do not have
        a SEQRES id will be dropped.

        Returns
        -------
        Chain
            Chain with seqres IDs as main index
        """
        return self._update_ids(
            self.residues.loc[:, "seqres_id"]
        )

    def filter_atoms(self, atom_name="CA"):
        """
        Filter coordinates of chain, e.g. to
        compute C_alpha-C_alpha distances

        Parameters
        ----------
        atom_name : str or list-like, optional (default: "CA")
            Name(s) of atoms to keep

        Returns
        -------
        Chain
            Chain containing only filtered atoms (and those
            residues that have such an atom)
        """
        if isinstance(atom_name, str):
            sel = self.coords.atom_name == atom_name
        else:
            sel = self.coords.atom_name.isin(atom_name)

        # update dataframe to rows having the right atom(s)
        coords = self.coords.loc[sel].copy()

        # reset atom index to consecutive numbers from 0
        coords = coords.reset_index(drop=True)

        # if there are residues without any atoms, remove
        # the entire residue
        residues = self.residues.loc[
            self.residues.index.isin(coords.residue_index)
        ].copy()

        return Chain(residues, coords)

    def filter_positions(self, positions):
        """
        Select a subset of positions from the chain
        
        Parameters
        ----------
        positions : list-like
            Set of residues that will be kept

        Returns
        -------
        Chain
            Chain containing only the selected residues
        """
        # map all positions to be strings
        positions = [str(p) for p in positions]

        residues = self.residues.loc[
            self.residues.id.isin(positions)
        ].copy()

        # drop coordinates of residues that were not kept
        coords = self.coords.loc[
            self.coords.residue_index.isin(residues.index)
        ]

        # reset atom index to consecutive numbers from 0
        coords = coords.reset_index(drop=True)

        return Chain(residues, coords)

    def remap(self, mapping, source_id="seqres_id"):
        """
        Remap chain into different numbering scheme
        (e.g. from seqres to uniprot numbering)

        Parameters
        ----------
        mapping : dict
            Mapping of residue identifiers from
            source_id (current main index of PDB chain)
            to new identifiers.

            mapping may either be:

            1. dict(str -> str) to map individual residue
               IDs. Keys and values of dictionary will be
               typecast to string before the mapping, so it
               is possible to pass in integer values too
               (if the source or target IDs are numbers)

            2. dict((int, int) -> (int, int)) to map ranges
               of numbers to ranges of numbers. This should
               typically be only used with RESSEQ or UniProt
               numbering. End index or range is \*inclusive*\
               Note that residue IDs in the end will still
               be handled as strings when mapping.

        source_id: {"seqres_id", "coord_id", "id"}, optional (default: "seqres_id")
            Residue identifier in chain to map \*from*\
            (will be used as key to access mapping)

        Returns
        -------
        Chain
            Chain with remapped numbering ("id" column
            in residues DataFrame)
        """
        # get one key to test which type of mapping we have
        # (range-based, or individual residues)
        test_key = next(iter(mapping.keys()))

        # test for range-based mapping
        if isinstance(test_key, Iterable) and not isinstance(test_key, str):
            # build up inidividual residue mapping
            final_mapping = {}
            for (source_start, source_end), (target_start, target_end) in mapping.items():
                source = map(
                    str, range(source_start, source_end + 1)
                )

                target = map(
                    str, range(target_start, target_end + 1)
                )

                final_mapping.update(
                    dict(zip(source, target))
                )
        else:
            # individual residue mapping, make sure all strings
            final_mapping = {
                str(s): str(t) for (s, t) in mapping.items()
            }

        # remap identifiers using mapping
        ids = self.residues.loc[:, source_id].map(
            final_mapping, na_action="ignore"
        )

        # create remapped chain
        return self._update_ids(ids)

    def to_file(self, fileobj, chain_id="A", end=True):
        """
        Write chain to a file in PDB format (mmCIF not yet
        supported).
        
        Note that PDB files written this function may not 
        be 100% compliant with the PDB format standards,
        in particular:

        * some HETATM records may turn into ATOM records
          when starting from an mmtf file, if the record
          has a one-letter code (such as MSE / M).

        * code does not print TER record at the end of
          a peptide chain
        
        Parameters
        ----------
        fileobj : file-like object
            Write to this file handle
        chain_id : str, optional (default: "A")
            Assign this chain name in file (allows to redefine
            chain name from whatever chain was originally)
        end : bool, optional (default: True)
            Print "END" record after chain (signals end of PDB file)
        """
        # merge residue-level information and atom-level information
        # in one joint table (i.e. the way data is presented in a
        # PDB/mmCIF file)
        x = self.coords.merge(
            self.residues, left_on="residue_index", right_index=True
        )

        # write one atom at a time
        for idx, r in x.iterrows():
            # split residue ID into position and insertion code
            cid = str(r["id"])
            if cid[-1].isalpha():
                coord_id = cid[:-1]
                ins_code = cid[-1]
            else:
                coord_id = cid
                ins_code = ""

            # atom element
            element = r["element"].upper()

            # need to split atom name into element and specifier
            # (e.g. beta carbon element:C, specifier:B) so we
            # can correctly justify in the 4-column atom
            # name field: first 2 (right-justified) are
            # element, second 2 (left-justified) are specifier
            src_atom_name = r["atom_name"]

            # to make things more complicated, there are cases like
            # HE21 (CNS) or 1HE2 (PDB) which break if assuming
            # that atom_element == element. In these cases, we
            # just use the full raw string
            if len(src_atom_name) == 4:
                atom_name = src_atom_name
            else:
                atom_element = src_atom_name[0:len(element)]
                atom_spec = src_atom_name[len(element):]
                atom_name = "{:>2s}{:<2s}".format(atom_element, atom_spec)

            # print charge if we have one (optional)
            charge = r["charge"]
            # test < and > to exclude nan values
            if charge < 0 or charge > 0:
                charge_sign = "-" if charge < 0 else "+"
                charge_value = abs(charge)
                charge_str = "{}{}".format(charge_value, charge_sign)
            else:
                charge_str = ""

            # format line and write
            s = PDB_FORMAT.format(
                atom="HETATM" if r["hetatm"] else "ATOM",
                atom_id=r["atom_id"],
                atom_name=atom_name,
                alt_loc_ind=r["alt_loc"],
                residue_name=r["three_letter_code"],
                chain_id=chain_id,
                residue_id=coord_id,
                ins_code=ins_code,
                x_coord=r["x"],
                y_coord=r["y"],
                z_coord=r["z"],
                occupancy=r["occupancy"],
                temp_factor=r["b_factor"],
                element_symbol=element,
                charge=charge_str,
            )
            fileobj.write(s + "\n")

        if end:
            fileobj.write("END" + 77 * " " + "\n")


class PDB:
    """
    Wrapper around PDB MMTF decoder object to access residue and
    coordinate information
    """
    def __init__(self, mmtf):
        """
        Initialize by further decoding information in mmtf object.

        Normally one should use from_file() and from_id() class
        methods to create object.

        Parameters
        ----------
        mmtf : mmtf.api.mmtf_reader.MMTFDecoder
            MMTF decoder object (as returned by fetch or parse
            function in mmtf module)
        """
        def _get_range(object_counts):
            """
            Extract index ranges for chains, residues and atoms
            """
            last_element = np.cumsum(object_counts, dtype=int)
            first_element = np.concatenate(
                (np.zeros(1, dtype=int), last_element[:-1])
            )

            return first_element, last_element

        # store raw MMTF decoder
        self.mmtf = mmtf

        # Step 1: summarize basic information about model
        # number of models in structure
        self.num_models = len(mmtf.chains_per_model)

        self.first_chain_index, self.last_chain_index = _get_range(
            mmtf.chains_per_model
        )

        # collect list which chain corresponds to which entity
        self.chain_to_entity = {}
        for i, ent in enumerate(mmtf.entity_list):
            for c in ent["chainIndexList"]:
                self.chain_to_entity[c] = i

        # Step 2: identify residues and corresponding atom indices

        # first/last index of residue (aka group) for each chain;
        # index these lists with index of chain
        self.first_residue_index, self.last_residue_index = _get_range(
            mmtf.groups_per_chain
        )

        # store explicit information about composition of residues
        def _group_info(field):
            return np.array(
                [x[field] for x in mmtf.group_list]
            )

        # three and one letter code names of different groups
        self._residue_names_3 = _group_info("groupName")
        self._residue_names_1 = _group_info("singleLetterCode")

        # atom types and overall number of atoms in each type of residue
        self._residue_type_atom_names = _group_info("atomNameList")
        self._residue_type_elements = _group_info("elementList")
        self._residue_type_charges = _group_info("formalChargeList")

        self._residue_type_num_atoms = np.array([len(x) for x in self._residue_type_atom_names])

        # prepare alternative location list as numpy array, with empty strings
        self.alt_loc_list = np.array(
            [x.replace("\x00", "") for x in mmtf.alt_loc_list]
        )

        # compute first and last atom index for each residue/group
        # (by fetching corresponding length for each group based on group type)
        self._residue_num_atoms = self._residue_type_num_atoms[mmtf.group_type_list]

        self.first_atom_index, self.last_atom_index = _get_range(
            self._residue_num_atoms
        )

        # assemble residue ID strings
        self.residue_ids = np.array([
            "{}{}".format(group_id, ins_code.replace("\x00", ""))
            for group_id, ins_code
            in zip(mmtf.group_id_list, mmtf.ins_code_list)
        ])

        # map secondary structure codes into DSSP symbols
        self.sec_struct = np.array(
            [MMTF_DSSP_CODE_MAP[x] for x in mmtf.sec_struct_list]
        )

    @classmethod
    def from_file(cls, filename):
        """
        Initialize structure from MMTF file

        Parameters
        ----------
        filename : str
            Path of MMTF file

        Returns
        -------
        PDB
            initialized PDB structure
        """
        try:
            return cls(parse(filename))
        except FileNotFoundError as e:
            raise ResourceError(
                "Could not find file {}".format(filename)
            ) from e

    @classmethod
    def from_id(cls, pdb_id):
        """
        Initialize structure by PDB ID (fetches
        structure from RCSB servers)

        Parameters
        ----------
        pdb_id : str
            PDB identifier (e.g. 1hzx)

        Returns
        -------
        PDB
            initialized PDB structure
        """
        try:
            return cls(fetch(pdb_id))
        except HTTPError as e:
            raise ResourceError(
                "Could not fetch MMTF data for {}".format(pdb_id)
            ) from e

    def get_chain(self, chain, model=0):
        """
        Extract residue information and atom coordinates
        for a given chain in PDB structure

        Parameters
        ----------
        chain : str
            Name of chain to be extracted (e.g. "A")
        model : int, optional (default: 0)
            Index of model to be extracted

        Returns
        -------
        Chain
            Chain object containing DataFrames listing residues
            and atom coordinates
        """
        if not (0 <= model < self.num_models):
            raise ValueError(
                "Illegal model index, can be from 0 up to {}".format(
                    self.num_models - 1
                )
            )

        # first and last index of chains corresponding to selected model
        first_chain_index = self.first_chain_index[model]
        last_chain_index = self.last_chain_index[model]

        # which model chains match our target PDB chain?
        chain_names = np.array(
            self.mmtf.chain_name_list[first_chain_index:last_chain_index]
        )

        # indices of chains that match chain name, in current model
        indices = np.arange(first_chain_index, last_chain_index, dtype=int)
        target_chain_indeces = indices[chain_names == chain]

        if len(target_chain_indeces) == 0:
            raise ValueError(
                "No chains with given name found"
            )

        # collect internal indeces of all residues/groups in chain
        residue_indeces = np.concatenate(
            np.array([
                np.arange(self.first_residue_index[i], self.last_residue_index[i])
                for i in target_chain_indeces
            ])
        )

        # chain indeces and identifiers for all residues
        # (not to be confused with chain name!);
        # maximum length 4 characters according to MMTF spec
        chain_indeces = np.concatenate([
            np.full(
                self.last_residue_index[i] - self.first_residue_index[i],
                i, dtype=int
            ) for i in target_chain_indeces
        ])

        chain_ids = np.array(self.mmtf.chain_id_list)[chain_indeces]

        # create dataframe representation of selected chain
        m = self.mmtf
        group_types = m.group_type_list[residue_indeces]

        res = OrderedDict([
            ("id", self.residue_ids[residue_indeces]),
            ("seqres_id", m.sequence_index_list[residue_indeces]),
            ("coord_id", self.residue_ids[residue_indeces]),
            ("one_letter_code", self._residue_names_1[group_types]),
            ("three_letter_code", self._residue_names_3[group_types]),
            ("chain_index", chain_indeces),
            ("chain_id", chain_ids),
            ("sec_struct", self.sec_struct[residue_indeces]),
        ])

        res_df = pd.DataFrame(res)

        # shift seqres indexing to start at 1;
        # However, do not add to positions without sequence index (-1)
        res_df.loc[res_df.seqres_id >= 0, "seqres_id"] += 1

        # turn all indeces into strings and create proper NaN values
        res_df.loc[:, "coord_id"] = (
            res_df.loc[:, "coord_id"].astype(str)
        )

        res_df.loc[:, "seqres_id"] = (
            res_df.loc[:, "seqres_id"].astype(str).replace("-1", np.nan)
        )
        # copy updated coordinate indeces
        res_df.loc[:, "id"] = res_df.loc[:, "coord_id"]

        res_df.loc[:, "one_letter_code"] = res_df.loc[:, "one_letter_code"].replace("?", np.nan)
        res_df.loc[:, "sec_struct"] = res_df.loc[:, "sec_struct"].replace("", np.nan)

        # reduction to 3-state secondary structure (following Rost & Sander)
        res_df.loc[:, "sec_struct_3state"] = res_df.loc[:, "sec_struct"].map(
            lambda x: DSSP_3_STATE_MAP[x], na_action="ignore"
        )

        # proxy for HETATM records - will not work e.g. for MSE,
        # which is listed as "M"
        res_df.loc[:, "hetatm"] = res_df.one_letter_code.isnull()

        # finally, get atom names and coordinates for all residues
        atom_first = self.first_atom_index[residue_indeces]
        atom_last = self.last_atom_index[residue_indeces]
        atom_names = np.concatenate(self._residue_type_atom_names[group_types])
        elements = np.concatenate(self._residue_type_elements[group_types])
        charges = np.concatenate(self._residue_type_charges[group_types])

        residue_number = np.repeat(res_df.index, atom_last - atom_first)
        atom_indices = np.concatenate([
            np.arange(self.first_atom_index[i], self.last_atom_index[i])
            for i in residue_indeces
        ])

        coords = OrderedDict([
            ("residue_index", residue_number),
            ("atom_id", self.mmtf.atom_id_list[atom_indices]),
            ("atom_name", atom_names),
            ("element", elements),
            ("charge", charges),
            ("x", self.mmtf.x_coord_list[atom_indices]),
            ("y", self.mmtf.y_coord_list[atom_indices]),
            ("z", self.mmtf.z_coord_list[atom_indices]),
            ("alt_loc", self.alt_loc_list[atom_indices]),
            ("occupancy", self.mmtf.occupancy_list[atom_indices]),
            ("b_factor", self.mmtf.b_factor_list[atom_indices]),
        ])

        coord_df = pd.DataFrame(coords)

        return Chain(res_df, coord_df)


class ClassicPDB:
    """
    Class to handle "classic" PDB and mmCIF formats
    (for new mmtf format see PDB class above). Wraps
    around Biopython PDB functionality to provide a consistent
    interface.
    
    Unlike the PDB class (based on mmtf), this object will
    not be able to extract SEQRES indices corresponding to
    ATOM-record residue indices.
    """
    def __init__(self, structure):
        """
        Initialize from Biopython Structure object

        Normally one should use from_file() class
        method to create object.

        Parameters
        ----------
        structure : Bio.PDB.Structure.Structure
            Biopython structure object (returned by
            PDBParser, MMCIFParser or MMTFParser)
        """
        self.structure = structure
        self.models = [
            m.get_id() for m in self.structure
        ]
        self.model_to_chains = {
            m: [
                c.get_id() for c in self.structure[m]
            ] for m in self.models
        }


    @classmethod
    def from_file(cls, filename, file_format="pdb"):
        """
        Initialize structure from PDB/mmCIF file

        Parameters
        ----------
        filename : str
            Path of file
        file_format : {"pdb", "cif"}, optional (default: "pdb")
            Format of structure (old PDB format or mmCIF)

        Returns
        -------
        ClassicPDB
            Initialized PDB structure
        """
        try:
            if file_format == "pdb":
                from Bio.PDB import PDBParser
                parser = PDBParser(QUIET=True)
            elif file_format == "cif":
                from Bio.PDB import FastMMCIFParser
                parser = FastMMCIFParser(QUIET=True)
            else:
                raise InvalidParameterError(
                    "Invalid file_format, valid options are: pdb, cif"
                )

            structure = parser.get_structure("", filename)
            return cls(structure)
        except FileNotFoundError as e:
            raise ResourceError(
                "Could not find file {}".format(filename)
            ) from e

    @classmethod
    def from_id(cls, pdb_id):
        """
        Initialize structure by PDB ID (fetches
        structure from RCSB servers)

        Parameters
        ----------
        pdb_id : str
            PDB identifier (e.g. 1hzx)

        Returns
        -------
        PDB
            initialized PDB structure
        """
        from urllib.error import URLError
        from Bio.PDB import PDBList
        pdblist = PDBList()

        try:
            # download PDB file to temporary directory
            pdb_file = pdblist.retrieve_pdb_file(pdb_id, pdir=tempdir())
            return cls.from_file(pdb_file, file_format="pdb")
        except URLError as e:
            raise ResourceError(
                "Could not fetch PDB data for {}".format(pdb_id)
            ) from e

    def get_chain(self, chain, model=0):
        """
        Extract residue information and atom coordinates
        for a given chain in PDB structure

        Parameters
        ----------
        chain : str
            Name of chain to be extracted (e.g. "A")
        model : int, optional (default: 0)
            Index of model to be extracted

        Returns
        -------
        Chain
            Chain object containing DataFrames listing residues
            and atom coordinates
        """
        # check requested model is valid
        if model not in self.models:
            raise ValueError(
                "Invalid model, valid models are: " +
                ",".join(map(str, self.models))
            )

        # check requested chain is valid
        if chain not in self.model_to_chains[model]:
            raise ValueError(
                "Invalid chain, valid chains are: " +
                ",".join(self.model_to_chains[model])
            )

        # get current chain
        c = self.structure[model][chain]

        res = DefaultOrderedDict(list)
        coords = DefaultOrderedDict(list)

        # iterate through all residues in chain and
        # accumulate information for dataframe building
        for r_idx, r in enumerate(c):
            het_flag, pos, ins_code = r.get_id()
            ins_code = ins_code.replace(" ", "")
            # print("x" + het_flag + "x", pos, "." + ins_code + ".")

            residue_id = "{}{}".format(pos, ins_code)
            res["id"].append(residue_id)
            # we don't have seqres ID from atom records, unless we
            # parsed additional annotation in header / mmCIF dict
            res["seqres_id"].append(np.nan)
            res["coord_id"].append(residue_id)

            # residue name in 3- and 1-letter code
            resname = r.get_resname()
            if resname in AA3_to_AA1:
                resname_one = AA3_to_AA1[resname]
            else:
                resname_one = np.nan
            res["one_letter_code"].append(resname_one)
            res["three_letter_code"].append(r.get_resname())

            # information that is only straightforward to get from mmtf
            res["chain_index"] = np.nan
            res["chain_id"] = np.nan
            res["sec_struct"] = np.nan
            res["sec_struct_3state"] = np.nan

            # at least we can identify easily what the hetatms
            # are (as opposed to mmtf)
            res["hetatm"].append(het_flag != " ")

            # now iterate through all atoms for current residue
            # and accumulate information; unpack atoms
            # with multiple locations (altloc)
            for a_idx, a in enumerate(r.get_unpacked_list()):
                # this index links residues to coords
                coords["residue_index"].append(r_idx)
                # atom-specific information
                coords["atom_id"].append(a.get_serial_number())
                coords["atom_name"].append(a.get_name())
                coords["element"].append(a.element)
                # cannot get charge information from Biopython?
                coords["charge"].append(np.nan)
                x, y, z = a.get_coord()
                coords["x"].append(x)
                coords["y"].append(y)
                coords["z"].append(z)
                coords["alt_loc"].append(
                    a.get_altloc().replace(" ", "")
                )
                coords["occupancy"].append(a.get_occupancy())
                coords["b_factor"].append(a.get_bfactor())

        # create residue table
        res_df = pd.DataFrame(res)

        # make sure that residue ids are strings
        res_df.loc[:, "coord_id"] = (
            res_df.loc[:, "coord_id"].astype(str)
        )

        # create coordinate table
        coord_df = pd.DataFrame(coords)

        return Chain(res_df, coord_df)


def load_structures(pdb_ids, structure_dir=None, raise_missing=True):
    """
    Load PDB structures from files / web

    Parameters
    ----------
    pdb_ids : Iterable
        List / iterable containing PDB identifiers
        to be loaded.
    structure_dir : str, optional (default: None)
        Path to directory with structures. Structures
        filenames must be in the format 5p21.mmtf.
        If a file can not be found, will try to fetch
        from web instead.
    raise_missing : bool, optional (default: True)
        Raise a ResourceError exception if any of the
        PDB IDs cannot be loaded. If False, missing
        entries will be ignored.

    Returns
    -------
    structures : dict(str -> PDB)
        Dictionary containing loaded structures.
        Keys (PDB identifiers) will be lower-case.

    Raises
    ------
    ResourceError
        Raised if raise_missing is True and any of the given
        PDB IDs cannot be loaded.
    """
    # collect loaded structures in dict(id -> PDB)
    structures = {}

    # load structure by structure
    for pdb_id in set(pdb_ids):
        pdb_id = pdb_id.lower()

        has_file = False
        if structure_dir is not None:
            structure_file = path.join(structure_dir, pdb_id + ".mmtf")
            has_file = valid_file(structure_file)

        try:
            # see if we can load locally from disk
            if has_file:
                structures[pdb_id] = PDB.from_file(structure_file)
            else:
                # otherwise fetch from web
                structures[pdb_id] = PDB.from_id(pdb_id)
        except (ResourceError, UnicodeDecodeError):
            # ResourceError: invalid PDB ID
            # UnicodeDecodeError: some random problem with mmtf library
            if raise_missing:
                raise

    return structures



from evcouplings.fold.tools import *
from evcouplings.fold.cns import *
from evcouplings.fold.filter import *
from evcouplings.fold.restraints import *
from evcouplings.fold.protocol import *
from evcouplings.fold.ranking import *



"""
Protocols for matching putatively interacting sequences
in protein complexes to create a concatenated sequence
alignment

Authors:
  Thomas A. Hopf
  Anna G. Green
"""

from evcouplings.couplings.mapping import Segment
from evcouplings.utils.config import (
    check_required, InvalidParameterError,
    read_config_file, write_config_file
)
from evcouplings.utils.system import (
    create_prefix_folders, valid_file,
    verify_resources,
)


def genome_distance(**kwargs):
    """
    Protocol:

    Concatenate alignments based on genomic distance

    Parameters
    ----------
    Mandatory kwargs arguments:
        See list below in code where calling check_required

        .. todo::

            Explain meaning of parameters in detail.

    Returns
    -------
    outcfg : dict
        Output configuration of the pipeline, including
        the following fields:

        .. todo::

            this is the full list normally returned by alignment protocol, decide which ones to keep.
            Mandatory:

            * alignment_file
            * focus_sequence
            * focus_mode
            * segments

        * alignment_file
        * [raw_alignment_file]
        * statistics_file
        * target_sequence_file
        * sequence_file
        * [annotation_file]
        * frequencies_file
        * identities_file
        * [hittable_file]
        * focus_mode
        * focus_sequence
        * segments
    """
    check_required(
        kwargs,
        [
            "prefix",
            "first_raw_focus_alignment_file", "second_raw_focus_alignment_file",
            "first_focus_sequence", "second_focus_sequence",
            "first_focus_mode", "second_focus_mode",
            "first_segments", "second_segments",
        ]
    )

    prefix = kwargs["prefix"]

    # make sure input alignments
    verify_resources(
        "Input alignment does not exist",
        kwargs["first_alignment_file"], kwargs["second_alignment_file"]
    )

    # make sure output directory exists
    create_prefix_folders(prefix)

    # -------------------------------------------------
    # TODO: implement concatenation functionality and
    # postprocessing functionality here
    # -------------------------------------------------

    def _modify_segments(seg_list, seg_prefix):
        # extract segments from list representation into objects
        segs = [
            Segment.from_list(s) for s in seg_list
        ]
        # update segment IDs
        for i, s in enumerate(segs, start=1):
            s.segment_id = "{}_{}".format(seg_prefix, i)

        return segs

    # merge segments - this allows to have more than one segment per
    # "monomer" alignment
    segments_1 = _modify_segments(kwargs["first_segments"], "A")
    segments_2 = _modify_segments(kwargs["second_segments"], "B")
    segments_complex = segments_1 + segments_2

    # make sure we return all the necessary information:
    # * alignment_file: final concatenated alignment that will go into plmc
    # * focus_sequence: this is the identifier of the concatenated target
    #   sequence which will be passed into plmc with -f

    outcfg = {
        "alignment_file": None,  # TODO: specify
        "focus_mode": True,
        "focus_sequence": None,  # TODO: specify
        "segments": [s.to_list() for s in segments_complex],
        # optional but good to have:
        "num_sites": None,
        "num_sequences": None,
        # "effective_sequences": n_eff # TODO: could compute this like in align stage
        # TODO: there are more outputs that we could add here (not mandatory),
        # e.g. single column frequencies in concatenated alignment
    }

    return outcfg


# list of available EC inference protocols
PROTOCOLS = {
    # concatenate based on genomic distance ("operon-based")
    "genome_distance": genome_distance,
}


def run(**kwargs):
    """
    Run alignment concatenation protocol

    Parameters
    ----------
    Mandatory kwargs arguments:
        protocol: concatenation protocol to run
        prefix: Output prefix for all generated files

    Returns
    -------
    outcfg : dict
        Output configuration of concatenation stage
        Dictionary with results in following fields (in brackets: not mandatory):

        .. todo::

            to be finalized after implementing protocols

        * alignment_file
        * focus_mode
        * focus_sequence
        * segments
        * num_sites
        * num_sequences
    """
    check_required(kwargs, ["protocol"])

    if kwargs["protocol"] not in PROTOCOLS:
        raise InvalidParameterError(
            "Invalid protocol selection: " +
            "{}. Valid protocols are: {}".format(
                kwargs["protocol"], ", ".join(PROTOCOLS.keys())
            )
        )

    return PROTOCOLS[kwargs["protocol"]](**kwargs)



"""
Compare evolutionary couplings to distances in 3D structures

Authors:
  Thomas A. Hopf
"""

import numpy as np


def add_distances(ec_table, dist_map, target_column="dist"):
    """
    Add pair distances to EC score table

    Parameters
    ----------
    ec_table : pandas.DataFrame
        List of evolutionary couplings, with pair
        positions in columns i and j
    dist_map : DistanceMap
        Distance map that will be used to annotate
        distances in ec_table
    target_column : str
        Name of column in which distances will be stored

    Returns
    -------
    pandas.DataFrame
        Couplings table with added distances
        in target_column. Pairs where no distance
        information is available will be np.nan
    """
    ec_table = ec_table.copy()

    ec_table.loc[:, target_column] = [
        dist_map.dist(i, j, raise_na=False)
        for i, j in zip(ec_table.i, ec_table.j)
    ]

    return ec_table


def add_precision(ec_table, dist_cutoff=5, score="cn",
                  min_sequence_dist=6, target_column="precision",
                  dist_column="dist"):
    """
    Compute precision of evolutionary couplings as predictor
    of 3D structure contacts

    Parameters
    ----------
    ec_table : pandas.DataFrame
        List of evolutionary couplings
    dist_cutoff : float, optional (default: 5)
        Upper distance cutoff (in Angstrom) for a
        pair to be considered a true positive contact
    score : str, optional (default: "cn")
        Column which contains coupling score. Table will
        be sorted in descending order by this score.
    min_sequence_dist : int, optional (default: 6)
        Minimal distance in primary sequence for an EC to
        be included in precision calculation
    target_column : str, optional (default: "precision")
        Name of column in which precision will be stored
    dist_column : str, optional (default: "dist")
        Name of column which contains pair distances

    Returns
    -------
    pandas.DataFrame
        EC table with added precision values as a
        function of EC rank (returned table will be
        sorted by score column)
    """
    # make sure list is sorted by score
    ec_table = ec_table.sort_values(by=score, ascending=False)

    if min_sequence_dist is not None:
        ec_table = ec_table.query("abs(i - j) >= @min_sequence_dist")

    ec_table = ec_table.copy()

    # number of true positive contacts
    true_pos_count = (ec_table.loc[:, dist_column] <= dist_cutoff).cumsum()

    # total number of contacts with specified distance
    pos_count = ec_table.loc[:, dist_column].notnull().cumsum()

    ec_table.loc[:, target_column] = true_pos_count / pos_count
    return ec_table


def coupling_scores_compared(ec_table, dist_map, dist_map_multimer=None,
                             dist_cutoff=5, output_file=None, score="cn",
                             min_sequence_dist=6):
    """
    Utility function to create "CouplingScores.csv"-style
    table

    Parameters
    ----------
    ec_table : pandas.DataFrame
        List of evolutionary couplings
    dist_map : DistanceMap
        Distance map that will be used to annotate
        distances in ec_table
    dist_map_multimer : DistanceMap, optional (default: None)
        Additional multimer distance map. If given,
        the distance for any EC pair will be the minimum
        out of the monomer and multimer distances.
    dist_cutoff : float, optional (default: 5)
        Upper distance cutoff (in Angstrom) for a
        pair to be considered a true positive contact
    output_file : str, optional (default: None)
        Store final table to this file
    score : str, optional (default: "cn")
        Column which contains coupling score. Table will
        be sorted in descending order by this score.
    min_sequence_dist : int, optional (default: 6)
        Minimal distance in primary sequence for an EC to
        be included in precision calculation

    Returns
    -------
    pandas.DataFrame
        EC table with added distances, and precision
        if dist_cutoff is given.
    """
    if dist_map_multimer is None:
        x = add_distances(ec_table, dist_map)
    else:
        x = add_distances(ec_table, dist_map, "dist_intra")
        x = add_distances(x, dist_map_multimer, "dist_multimer")
        x.loc[:, "dist"] = np.fmin(
            x.dist_intra, x.dist_multimer
        )

    if min_sequence_dist is not None:
        x = x.query("abs(i - j) >= @min_sequence_dist")

    # if distance cutoff is given, add precision
    if dist_cutoff is not None:
        x = add_precision(
            x, dist_cutoff, score=score,
            min_sequence_dist=min_sequence_dist
        )

    # also save to file if path is given
    if output_file is not None:
        x.to_csv(output_file, index=False)

    return x



"""
Job status database models and handling

Authors:
  Thomas A. Hopf
"""

from sqlalchemy import (
    Column, Integer, String, DateTime,
    create_engine, func
)
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

# enumeration of possible job status values
EStatus = (lambda **enums: type('Enum', (), enums))(
    INIT="initialized",
    PEND="pending",
    RUN="running",
    DONE="done",
    FAIL="failed",  # job failed due to bug
    TERM="terminated",  # job was terminated externally
)

Base = declarative_base()


class ComputeJob(Base):
    """
    Single compute job. Holds general information about job
    and its status, but not about individual parameters
    (these are stored in config file to keep table schema
    stable).
    """
    __tablename__ = "runs"

    # unique ID of this single compute job
    id = Column(Integer, primary_key=True)

    # human-readable job name (must be unique)
    name = Column(String(100), unique=True)

    # job prefix
    prefix = Column(String(1024))

    # path to configuration file used for job
    config_file = Column(String(1024))

    # location - not used right now, but could
    # be used to point to cloud locations, etc.
    location = Column(String(1024))

    # job group this job is associated with
    # (foreign key in group key if this table
    # is present, e.g. in webserver).
    group_id = Column(Integer)

    # job status ("pending", "running", "finished",
    # "failed", "terminated")
    status = Column(String(50))

    # stage of computational pipeline
    # ("align", "couplings", ...)
    stage = Column(String(50))

    # time the job started running
    time_started = Column(DateTime())

    # time the job finished running
    time_finished = Column(DateTime())


def update_job_status(config, status=None, stage=None):
    """
    Update job status based on configuration and
    update request by pipeline

    Parameters
    ----------
    config : dict-like
        Job configuration dictionary. Accessed entries
        are {
            global : {prefix}
            management: {database_uri, job_name}
        }
    status : str, optional (default: None)
        If not None, update job status to this value
    stage : str, optional (default: None)
        If not None, update job stage to this value
    """
    # extract database information from job configuration
    # (database URI and job id), as well as job prefix
    mgmt = config.get("management", {})
    uri = mgmt.get("database_uri", None)
    job_name = mgmt.get("job_name", None)
    prefix = config.get("global", {}).get("prefix", None)

    # if we don't have these settings, cannot update job status
    if uri is None or job_name is None:
        return

    # connect to DB and create session
    engine = create_engine(uri)
    Session = sessionmaker(bind=engine)
    session = Session()

    # make sure all tables are there in database
    Base.metadata.create_all(bind=engine)

    # see if we can find the job in the database already
    q = session.query(ComputeJob).filter_by(name=job_name)
    num_rows = len(q.all())

    # create new entry if not already existing
    if num_rows == 0:
        r = ComputeJob(
            name=job_name,
            prefix=prefix,
            status="pending",
        )
        session.add(r)
        session.commit()
    else:
        # can only be one row due to unique constraint
        r = q.one()

    # if status is given, update
    if status is not None:
        r.status = status

    # if stage is given, update
    if stage is not None:
        r.stage = stage

    # if start time has not been set yet, do so
    if r.time_started is None:
        r.time_started = func.now()

    # update finish time (i.e. final finish
    # time when job status is set for the last time)
    r.time_finished = func.now()

    # commit changes to database
    session.commit()

    # TODO: remove eventually
    y = [x.__dict__ for x in session.query(ComputeJob).filter_by(name=job_name).all()]
    print()
    print(y)
    print()
    # print(y["status"], y["stage"])
    # print(json.dumps(y[0], indent=2))
    # TODO: remove

    # close session again
    session.close()



"""
Class and functions for reading, storing and manipulating
multiple sequence alignments.

Authors:
  Thomas A. Hopf
"""

from collections import namedtuple, OrderedDict, defaultdict
from copy import deepcopy
import re

import numpy as np
from numba import jit

from evcouplings.utils.calculations import entropy
from evcouplings.utils.helpers import DefaultOrderedDict, wrap

# constants
GAP = "-"
MATCH_GAP = GAP
INSERT_GAP = "."

ALPHABET_PROTEIN_NOGAP = "ACDEFGHIKLMNPQRSTVWY"
ALPHABET_PROTEIN = GAP + ALPHABET_PROTEIN_NOGAP

ALPHABET_DNA_NOGAP = "ACGT"
ALPHABET_DNA = GAP + ALPHABET_DNA_NOGAP

ALPHABET_RNA_NOGAP = "ACGU"
ALPHABET_RNA = GAP + ALPHABET_RNA_NOGAP


def read_fasta(fileobj):
    """
    Generator function to read a FASTA-format file
    (includes aligned FASTA, A2M, A3M formats)

    Parameters
    ----------
    fileobj : file-like object
        FASTA alignment file

    Returns
    -------
    generator of (str, str) tuples
        Returns tuples of (sequence ID, sequence)
    """
    current_sequence = ""
    current_id = None

    for line in fileobj:
        # Start reading new entry. If we already have
        # seen an entry before, return it first.
        if line.startswith(">"):
            if current_id is not None:
                yield current_id, current_sequence

            current_id = line.rstrip()[1:]
            current_sequence = ""

        elif not line.startswith(";"):
            current_sequence += line.rstrip()

    # Also do not forget last entry in file
    yield current_id, current_sequence


def write_fasta(sequences, fileobj, width=80):
    """
    Write a list of IDs/sequences to a FASTA-format file

    Parameters
    ----------
    sequences : Iterable
        Iterator returning tuples(str, str), where
        first value is header/ID and second the
        sequence
    fileobj : file-like obj
        File to which alignment will be written
    """
    for (seq_id, seq) in sequences:
        fileobj.write(">{}\n".format(seq_id))
        fileobj.write(wrap(seq, width=width) + "\n")


def write_aln(sequences, fileobj, width=80):
    """
    Write a list of IDs/sequences to a ALN-format file

    Currently, the file will not contain headers but simply
    a block matrix of the alignment

    Parameters
    ----------
    sequences : Iterable
        Iterator returning tuples(str, str), where
        first value is header/ID and second the
        sequence
    fileobj : file-like obj
        File to which alignment will be written
    """
    for (seq_id, seq) in sequences:
        fileobj.write(seq + "\n")


# Holds information of a parsed Stockholm alignment file
StockholmAlignment = namedtuple(
    "StockholmAlignment",
    ["seqs", "gf", "gc", "gs", "gr"]
)


def read_stockholm(fileobj, read_annotation=False):
    """
    Generator function to read Stockholm format alignment
    file (e.g. from hmmer).

    .. note::

        Generator iterates over different alignments in the
        same file (but not over individual sequences, which makes
        little sense due to wrapped Stockholm alignments).


    Parameters
    ----------
    fileobj : file-like object
        Stockholm alignment file
    read_annotation : bool, optional (default=False)
        Read annotation columns from alignment

    Returns
    -------
    StockholmAlignment
        namedtuple with the following fields:
        seqs, gf, gc, gs, gr (all DefaultOrderedDict)

    Raises
    ------
    ValueError
    """
    seqs = DefaultOrderedDict(str)

    """
    Markup definition: http://sonnhammer.sbc.su.se/Stockholm.html
    #=GF <feature> <Generic per-File annotation, free text>
    #=GC <feature> <Generic per-Column annotation, exactly 1 char per column>
    #=GS <seqname> <feature> <Generic per-Sequence annotation, free text>
    #=GR <seqname> <feature> <Generic per-Residue annotation, exactly 1 char per residue>
    """
    gf = DefaultOrderedDict(list)
    gc = DefaultOrderedDict(str)
    gs = DefaultOrderedDict(lambda: DefaultOrderedDict(list))
    gr = DefaultOrderedDict(lambda: DefaultOrderedDict(str))

    # line counter within current alignment (can be more than one per file)
    i = 0

    # read alignment
    for line in fileobj:
        if i == 0 and not line.startswith("# STOCKHOLM 1.0"):
            raise ValueError(
                "Not a valid Stockholm alignment: "
                "Header missing. {}".format(line.rstrip())
            )

        # annotation lines
        if line.startswith("#"):
            if read_annotation:
                if line.startswith("#=GF"):
                    # can have multiple lines for the same feature
                    _, feat, val = line.rstrip().split(maxsplit=2)
                    gf[feat].append(val)
                elif line.startswith("#=GC"):
                    # only one line with the same GC label
                    _, feat, seq = line.rstrip().split(maxsplit=2)
                    gc[feat] += seq
                elif line.startswith("#=GS"):
                    # can have multiple lines for the same feature
                    _, seq_id, feat, val = line.rstrip().split(maxsplit=3)
                    gs[seq_id][feat] = val
                elif line.startswith("#=GR"):
                    # per sequence, only one line with a certain GR feature
                    _, seq_id, feat, seq = line.rstrip().split()
                    gr[seq_id][feat] += seq

            i += 1

        # terminator line for current alignment;
        # only yield once we see // to avoid reading
        # truncated alignments
        elif line.startswith("//"):
            yield StockholmAlignment(seqs, gf, gc, gs, gr)

            # reset counter to check for valid start of alignment
            # once we read the next line
            i = 0

        # actual alignment lines
        else:
            splitted = line.rstrip().split(maxsplit=2)
            # there might be empty lines, so check for valid split
            if len(splitted) == 2:
                seq_id, seq = splitted
                seqs[seq_id] += seq

            i += 1

    # Do NOT yield at the end without // to avoid returning truncated alignments


def read_a3m(fileobj, inserts="first"):
    """
    Read an alignment in compressed a3m format and expand
    into a2m format.

    .. note::

        this function is currently not able to keep inserts in all the sequences

    ..todo::

        implement this

    Parameters
    ----------
    fileobj : file-like object
        A3M alignment file
    inserts : {"first", "delete"}
        Keep inserts in first sequence, or delete
        any insert column and keep only match state
        columns.

    Returns
    -------
    OrderedDict
        Sequences in alignment (key: ID, value: sequence),
        in order they appeared in input file

    Raises
    ------
    ValueError
        Upon invalid choice of insert strategy
    """
    seqs = OrderedDict()

    for i, (seq_id, seq) in enumerate(read_fasta(fileobj)):
        # remove any insert gaps that may still be in alignment
        # (just to be sure)
        seq = seq.replace(".", "")

        if inserts == "first":
            # define "spacing" of uppercase columns in
            # final alignment based on target sequence;
            # remaining columns will be filled with insert
            # gaps in the other sequences
            if i == 0:
                uppercase_cols = [
                    j for (j, c) in enumerate(seq)
                    if (c == c.upper() or c == "-")
                ]
                gap_template = np.array(["."] * len(seq))
                filled_seq = seq
            else:
                uppercase_chars = [
                    c for c in seq if c == c.upper() or c == "-"
                ]
                filled = np.copy(gap_template)
                filled[uppercase_cols] = uppercase_chars
                filled_seq = "".join(filled)

        elif inserts == "delete":
            # remove all lowercase letters and insert gaps .;
            # since each sequence must have same number of
            # uppercase letters or match gaps -, this gives
            # the final sequence in alignment
            seq = "".join([c for c in seq if c == c.upper() and c != "."])
        else:
            raise ValueError(
                "Invalid option for inserts: {}".format(inserts)
            )

        seqs[seq_id] = filled_seq

    return seqs


def write_a3m(sequences, fileobj, insert_gap=INSERT_GAP, width=80):
    """
    Write a list of IDs/sequences to a FASTA-format file

    Parameters
    ----------
    sequences : Iterable
        Iterator returning tuples(str, str), where
        first value is header/ID and second the
        sequence
    fileobj : file-like obj
        File to which alignment will be written
    """
    for (seq_id, seq) in sequences:
        fileobj.write(">{}\n".format(seq_id))
        fileobj.write(seq.replace(insert_gap, "") + "\n")


def detect_format(fileobj):
    """
    Detect if an alignment file is in FASTA or
    Stockholm format.

    Parameters
    ----------
    fileobj : file-like obj
        Alignment file for which to detect format

    Returns
    -------
    format : {"fasta", "stockholm", None}
        Format of alignment, None if not detectable
    """
    for i, line in enumerate(fileobj):
        # must be first line of Stockholm file by definition
        if i == 0 and line.startswith("# STOCKHOLM 1.0"):
            return "stockholm"

        # This indicates a FASTA file
        if line.startswith(">"):
            return "fasta"

        # Skip comment lines and empty lines for FASTA detection
        if line.startswith(";") or line.rstrip() == "":
            continue

        # Arriving here means we could not detect format
        return None


def parse_header(header):
    """
    Extract ID of the (overall) sequence and the
    sequence range fromat a sequence header of the form
    sequenceID/start-end.

    If the header contains any additional information
    after the first whitespace character (e.g. sequence annotation),
    it will be discarded before parsing. If there is no sequence
    range, only the id (part of string before whitespace) will
    be returned but no range.

    Parameters
    ----------
    header : str
        Sequence header

    Returns
    -------
    seq_id : str
        Sequence identifier
    start : int
        Start of sequence range. Will be None if it cannot
        be extracted from the header.
    stop : int
        End of sequence range. Will be None if it cannot
        be extracted from the header.
    """
    # discard anything in header that might come after the
    # first whitespace (by convention this is typically annotation)
    header = header.split()[0]

    # try to extract region from sequence header
    m = re.search("(.+)/(\d+)-(\d+)", header)
    if m:
        id_, start_str, end_str = m.groups()
        region_start, region_end = int(start_str), int(end_str)
        return id_, region_start, region_end
    else:
        # cannot find region, so just give back sequence iD
        return header, None, None


def sequences_to_matrix(sequences):
    """
    Transforms a list of sequences into a
    numpy array.

    Parameters
    ----------
    sequences : list-like (str)
        List of strings containing aligned sequences

    Returns
    -------
    numpy.array
        2D array containing sequence alignment
        (first axis: sequences, second axis: columns)
    """
    if len(sequences) == 0:
        raise ValueError("Need at least one sequence")

    N = len(sequences)
    L = len(next(iter(sequences)))
    matrix = np.empty((N, L), dtype=np.str)

    for i, seq in enumerate(sequences):
        if len(seq) != L:
            raise ValueError(
                "Sequences have differing lengths: i={} L_0={} L_i={}".format(
                    i, L, len(seq)
                )
            )

        matrix[i] = np.array(list(seq))

    return matrix


def map_from_alphabet(alphabet=ALPHABET_PROTEIN, default=GAP):
    """
    Creates a mapping dictionary from a given alphabet.

    Parameters
    ----------
    alphabet : str
        Alphabet for remapping. Elements will
        be remapped according to alphabet starting
        from 0
    default : Elements in matrix that are not
        contained in alphabet will be treated as
        this character

    Raises
    ------
    ValueError
        For invalid default character
    """
    map_ = {
        c: i for i, c in enumerate(alphabet)
    }

    try:
        default = map_[default]
    except KeyError:
        raise ValueError(
            "Default {} is not in alphabet {}".format(default, alphabet)
        )

    return defaultdict(lambda: default, map_)


def map_matrix(matrix, map_):
    """
    Map elements in a numpy array using alphabet

    Parameters
    ----------
    matrix : np.array
        Matrix that should be remapped
    map_ : defaultdict
        Map that will be applied to matrix elements

    Returns
    -------
    np.array
        Remapped matrix
    """
    return np.vectorize(map_.__getitem__)(matrix)


class Alignment:
    """
    Container to store and manipulate multiple sequence alignments.

    .. note::

        Important:

        1. Sequence annotation currently is not transformed when
           selecting subsets of columns or positions (e.g. affects GR and GC
           lines in Stockholm alignments)
        2. Sequence ranges in IDs are not adjusted when selecting
           subsets of positions
    """
    def __init__(self, sequence_matrix, sequence_ids=None, annotation=None,
                 alphabet=ALPHABET_PROTEIN):
        """
        Create new alignment object from ready-made components.

        .. note::

            Use factory method Alignment.from_file to create alignment from file,
            or Alignment.from_dict from dictionary of sequences.

        Parameters
        ----------
        sequence_matrix : np.array
            N x L array of characters in the alignment
            (N=number of sequences, L=width of alignment)
        sequence_ids : list-like, optional (default=None)
            Sequence names of alignment members (must have N elements).
            If None, defaults sequence IDs to "0", "1", ...
        annotation : dict-like
            Annotation for sequence alignment

        Raises
        ------
        ValueError
            If dimensions of sequence_matrix and sequence_ids
            are inconsistent
        """
        self.matrix = np.array(sequence_matrix)
        self.N, self.L = self.matrix.shape

        # characters coding for gaps in match-state and insert
        # columns of the alignment
        self._match_gap = MATCH_GAP
        self._insert_gap = INSERT_GAP

        # defined alphabet of alignment
        self.alphabet = alphabet
        self.alphabet_default = self._match_gap

        self.alphabet_map = map_from_alphabet(
            self.alphabet, default=self.alphabet_default
        )
        self.num_symbols = len(self.alphabet_map)

        # Alignment matrix remapped into in integers
        # Will only be calculated if necessary for downstream
        # calculations
        self.matrix_mapped = None
        self.num_cluster_members = None
        self.weights = None
        self._frequencies = None
        self._pair_frequencies = None

        if sequence_ids is None:
            # default to numbering sequences if not given
            self.ids = [str(i) for i in range(self.N)]
        else:
            if len(sequence_ids) != self.N:
                raise ValueError(
                    "Number of sequence IDs and length of "
                    "alignment do not match".format(
                        len(sequence_ids), self.L
                    )
                )

            # make sure we get rid of iterators etc.
            self.ids = np.array(list(sequence_ids))

        self.id_to_index = {
            id_: i for i, id_ in enumerate(self.ids)
        }

        if annotation is not None:
            self.annotation = annotation
        else:
            self.annotation = {}

    @classmethod
    def from_dict(cls, sequences, **kwargs):
        """
        Construct an alignment object from a dictionary
        with sequence IDs as keys and aligned sequences
        as values.

        Parameters
        ----------
        sequences : dict-like
            Dictionary with pairs of sequence ID (key) and
            aligned sequence (value)

        Returns
        -------
        Alignment
            initialized alignment
        """
        matrix = sequences_to_matrix(sequences.values())

        return cls(
            matrix, sequences.keys(), **kwargs
        )

    @classmethod
    def from_file(cls, fileobj, format="fasta",
                  a3m_inserts="first", **kwargs):
        """
        Construct an alignment object by reading in an
        alignment file.

        Parameters
        ----------
        fileobj : file-like obj
            Alignment to be read in
        format : {"fasta", "stockholm", "a3m"}
            Format of input alignment
        a3m_inserts : {"first", "delete"}, optional (default: "first")
            Strategy to deal with inserts in a3m alignment files
            (see read_a3m documentation for details)

        Returns
        -------
        Alignment
            Parsed alignment

        Raises
        ------
        ValueError
            For invalid alignments or alignment formats
        """
        annotation = {}
        # read in sequence alignment from file

        if format == "fasta":
            seqs = OrderedDict()
            for seq_id, seq in read_fasta(fileobj):
                seqs[seq_id] = seq
        elif format == "stockholm":
            # only reads first Stockholm alignment contained in file
            ali = next(read_stockholm(fileobj, read_annotation=True))
            seqs = ali.seqs
            annotation["GF"] = ali.gf
            annotation["GC"] = ali.gc
            annotation["GS"] = ali.gs
            annotation["GR"] = ali.gr
            kwargs["annotation"] = annotation
        elif format == "a3m":
            seqs = read_a3m(fileobj, inserts=a3m_inserts)
        else:
            raise ValueError("Invalid alignment format: {}".format(format))

        return cls.from_dict(seqs, **kwargs)

    def __getitem__(self, index):
        """
        .. todo::

            eventually this should allow fancy indexing and offer the functionality of select()
        """
        if index in self.id_to_index:
            return self.matrix[self.id_to_index[index], :]
        elif index in range(self.N):
            return self.matrix[index, :]
        else:
            raise KeyError(
                "Not a valid index for sequence alignment: {}".format(index)
            )

    def __len__(self):
        return self.N

    def count(self, char, axis="pos", normalize=True):
        """
        Count occurrences of a character in the sequence
        alignment.

        .. note::

            The counts are raw counts not adjusted for
            sequence redundancy.

        Parameters
        ----------
        char : str
            Character which is counted
        axis : {"pos", "seq"}, optional (default="pos")
            Count along positions or sequences
        normalize : bool, optional (default=True)
            Normalize count for length of axis (i.e. relative count)

        Returns
        -------
        np.array
            Vector containing counts of char along the axis

        Raises
        ------
        ValueError
            Upon invalid axis specification
        """
        if axis == "pos":
            naxis = 0
        elif axis == "seq":
            naxis = 1
        else:
            raise ValueError("Invalid axis: {}".format(axis))

        c = np.sum(self.matrix == char, axis=naxis)
        if normalize:
            c = c / self.matrix.shape[naxis]

        return c

    def select(self, columns=None, sequences=None):
        """
        Create a sub-alignment that contains a subset of
        sequences and/or columns.

        .. note::

            This does currently not adjust the indices
            of the sequences. Annotation in the original alignment
            will be lost and not passed on to the new object.

        Parameters
        ----------
        columns : np.array(bool) or np.array(int), optional
            Vector containing True for each column that
            should be retained, False otherwise; or the
            indices of columns that should be selected
        sequences : np.array(bool) or np.array(int), optional
            Vector containing True for each sequence that
            should be retained, False otherwise; or the
            indices of sequences that should be selected

        Returns
        -------
        Alignment
            Alignment with selected columns and sequences
            (note this alignment looses annotation)
        """
        if columns is None and sequences is None:
            return self

        sel_matrix = self.matrix
        ids = self.ids

        if columns is not None:
            sel_matrix = sel_matrix[:, columns]

        if sequences is not None:
            sel_matrix = sel_matrix[sequences, :]
            ids = ids[sequences]

        # do not copy annotation since it may become
        # inconsistent
        return Alignment(
            np.copy(sel_matrix), np.copy(ids),
            alphabet=self.alphabet
        )

    def apply(self, columns=None, sequences=None, func=np.char.lower):
        """
        Apply a function along columns and/or rows of alignment matrix,
        or to entire matrix.

        Parameters
        ----------
        columns : np.array(bool) or np.array(int), optional
            Vector containing True for each column that
            should be retained, False otherwise; or the
            indices of columns that should be selected
        sequences : np.array(bool) or np.array(int), optional
            Vector containing True for each sequence that
            should be retained, False otherwise; or the
            indices of sequences that should be selected
        func : callable
            Vectorized numpy function that will be applied to
            the selected subset of the alignment matrix

        Returns
        -------
        Alignment
            Alignment with modified columns and sequences
            (this alignment maintains annotation)
        """
        mod_matrix = np.copy(self.matrix)

        if columns is None and sequences is None:
            return self
        else:
            if columns is not None:
                mod_matrix[:, columns] = func(mod_matrix[:, columns])

            if sequences is not None:
                mod_matrix[sequences, :] = func(mod_matrix[sequences, :])

        return Alignment(
            mod_matrix, np.copy(self.ids), deepcopy(self.annotation),
            alphabet=self.alphabet
        )

    def replace(self, original, replacement, columns=None, sequences=None):
        """
        Replace character with another in full matrix or
        subset of columns/sequences.

        Parameters
        ----------
        original : char
            Character that should be replaced
        replacement : char
            Replacement character
        columns : numpy index array
            See self.apply for explanation
        sequences : numpy index array
            See self.apply for explanation

        Returns
        -------
        Alignment
            Alignment with replaced characters

        """
        return self.apply(
            columns, sequences,
            func=lambda x: np.char.replace(
                x, original, replacement
            )
        )

    def lowercase_columns(self, columns):
        """
        Change a subset of columns to lowercase character
        and replace "-" gaps with "." gaps, e.g. to exclude
        them from EC calculations

        Parameters
        ----------
        columns : numpy index array
            Subset of columns to make lowercase

        Returns
        -------
        Alignment
            Alignment with lowercase columns
        """
        return self.apply(
            columns=columns, func=np.char.lower
        ).replace(
            self._match_gap, self._insert_gap, columns=columns
        )

    def __ensure_mapped_matrix(self):
        """
        Ensure self.matrix_mapped exists
        """
        if self.matrix_mapped is None:
            self.matrix_mapped = map_matrix(
                self.matrix, self.alphabet_map
            )

    def set_weights(self, identity_threshold=0.8):
        """
        Calculate weights for sequences in alignment by
        clustering all sequences with sequence identity
        greater or equal to the given threshold.

        .. note::

            This method sets self.weights. After this method was called, methods/attributes such as
            self.frequencies or self.conservation()
            will make use of sequence weights.

        .. note::

            (Implementation: cannot use property here since we need identity threshold as a parameter....)

        Parameters
        ----------
        identity_threshold : float, optional (default: 0.8)
            Sequence identity threshold
        """
        self.__ensure_mapped_matrix()

        self.num_cluster_members = num_cluster_members(
            self.matrix_mapped, identity_threshold
        )
        self.weights = 1.0 / self.num_cluster_members

        # reset frequencies, since these were based on
        # different weights before or had no weights at all
        self._frequencies = None
        self._pair_frequencies = None

    @property
    def frequencies(self):
        """
        Returns/calculates single-site frequencies of symbols in alignment.
        Also sets self._frequencies member variable for later reuse.

        Previously calculated sequence weights using self.set_weights()
        will be used to adjust frequency counts; otherwise, each sequence
        will contribute with equal weight.

        Returns
        -------
        np.array
            Reference to self._frequencies
        """
        if self._frequencies is None:
            self.__ensure_mapped_matrix()

            # use precalculated sequence weights, but only
            # if we have explicitly calculated them before
            # (expensive calculation)
            if self.weights is None:
                weights = np.ones((self.N))
            else:
                weights = self.weights

            self._frequencies = frequencies(
                self.matrix_mapped, weights, self.num_symbols
            )

        return self._frequencies

    @property
    def pair_frequencies(self):
        """
        Returns/calculates pairwise frequencies of symbols in alignment.
        Also sets self._pair_frequencies member variable for later reuse.

        Previously calculated sequence weights using self.set_weights()
        will be used to adjust frequency counts; otherwise, each sequence
        will contribute with equal weight.

        Returns
        -------
        np.array
            Reference to self._pair_frequencies
        """
        if self._pair_frequencies is None:
            self.__ensure_mapped_matrix()

            if self.weights is None:
                weights = np.ones((self.N))
            else:
                weights = self.weights

            self._pair_frequencies = pair_frequencies(
                self.matrix_mapped, weights,
                self.num_symbols, self.frequencies
            )

        return self._pair_frequencies

    def identities_to(self, seq, normalize=True):
        """
        Calculate sequence identity between sequence
        and all sequences in the alignment.

        seq : np.array, list-like, or str
            Sequence for comparison
        normalize : bool, optional (default: True)
            Calculate relative identity between 0 and 1
            by normalizing with length of alignment
        """
        self.__ensure_mapped_matrix()

        # make sure this doesnt break with strings
        seq = np.array(list(seq))

        seq_mapped = map_matrix(seq, self.alphabet_map)
        ids = identities_to_seq(seq_mapped, self.matrix_mapped)

        if normalize:
            return ids / self.L
        else:
            return ids

    def conservation(self, normalize=True):
        """
        Calculate per-column conservation of sequence alignment
        based on entropy of single-column frequency distribution.

        If self.set_weights() was called previously, the
        frequencies used to calculate site entropies will be based
        on reweighted sequences.

        Parameters
        ----------
        normalize : bool, optional (default: True)
            Transform column entropy to range 0 (no conservation)
            to 1 (fully conserved)

        Returns
        -------
        np.array
            Vector of length L with conservation scores
        """
        return np.apply_along_axis(
            lambda x: entropy(x, normalize=normalize),
            axis=1, arr=self.frequencies
        )

    def write(self, fileobj, format="fasta", width=80):
        """
        Write an alignment to a file.

        Parameters
        ----------
        fileobj : file-like object
            File to which alignment is saved
        format : {"fasta", "aln", "a3m"}
            Output format for alignment
        width : int
            Column width for fasta alignment

        Raises
        ------
        ValueError
            Upon invalid file format specification
        """
        seqs = (
            (id_, "".join(self.matrix[i]))
            for (i, id_) in enumerate(self.ids)
        )

        if format == "fasta":
            write_fasta(seqs, fileobj, width)
        elif format == "a3m":
            write_a3m(seqs, fileobj, self._insert_gap, width)
        elif format == "aln":
            write_aln(seqs, fileobj, width)
        else:
            raise ValueError(
                "Invalid alignment format: {}".format(format)
            )


@jit(nopython=True)
def frequencies(matrix, seq_weights, num_symbols):
    """
    Calculate single-site frequencies of symbols in alignment

    Parameters
    ----------
    matrix : np.array
        N x L matrix containing N sequences of length L.
        Matrix must be mapped to range(0, num_symbols) using
        map_matrix function
    seq_weights : np.array
        Vector of length N containing weight for each sequence
    num_symbols : int
        Number of different symbols contained in alignment

    Returns
    -------
    np.array
        Matrix of size L x num_symbols containing relative
        column frequencies of all characters
    """
    N, L = matrix.shape
    fi = np.zeros((L, num_symbols))
    for s in range(N):
        for i in range(L):
            fi[i, matrix[s, i]] += seq_weights[s]

    return fi / seq_weights.sum()


@jit(nopython=True)
def pair_frequencies(matrix, seq_weights, num_symbols, fi):
    """
    Calculate pairwise frequencies of symbols in alignment.

    Parameters
    ----------
    matrix : np.array
        N x L matrix containing N sequences of length L.
        Matrix must be mapped to range(0, num_symbols) using
        map_matrix function
    seq_weights : np.array
        Vector of length N containing weight for each sequence
    num_symbols : int
        Number of different symbols contained in alignment
    fi : np.array
        Matrix of size L x num_symbols containing relative
        column frequencies of all characters.

    Returns
    -------
    np.array
        Matrix of size L x L x num_symbols x num_symbols containing
        relative pairwise frequencies of all character combinations
    """
    N, L = matrix.shape
    fij = np.zeros((L, L, num_symbols, num_symbols))
    for s in range(N):
        for i in range(L):
            for j in range(i + 1, L):
                fij[i, j, matrix[s, i], matrix[s, j]] += seq_weights[s]
                fij[j, i, matrix[s, j], matrix[s, i]] = fij[i, j, matrix[s, i], matrix[s, j]]

    # normalize frequencies by the number
    # of effective sequences
    fij /= seq_weights.sum()

    # set the frequency of a pair (alpha, alpha)
    # in position i to the respective single-site
    # frequency of alpha in position i
    for i in range(L):
        for alpha in range(num_symbols):
            fij[i, i, alpha, alpha] = fi[i, alpha]

    return fij


@jit(nopython=True)
def identities_to_seq(seq, matrix):
    """
    Calculate number of identities to given target sequence
    for all sequences in the matrix

    Parameters
    ----------
    seq : np.array
        Vector of length L containing mapped sequence
        (using map_matrix function)
    matrix : np.array
        N x L matrix containing N sequences of length L.
        Matrix must be mapped to range(0, num_symbols)
        using map_matrix function

    Returns
    -------
    np.array
        Vector of length N containing number of identities
        to each sequence in matrix
    """
    N, L = matrix.shape
    identities = np.zeros((N, ))

    for i in range(N):
        id_i = 0
        for j in range(L):
            if matrix[i, j] == seq[j]:
                id_i += 1

        identities[i] = id_i

    return identities


@jit(nopython=True)
def num_cluster_members(matrix, identity_threshold):
    """
    Calculate number of sequences in alignment
    within given identity_threshold of each other

    Parameters
    ----------
    matrix : np.array
        N x L matrix containing N sequences of length L.
        Matrix must be mapped to range(0, num_symbols) using
        map_matrix function
    identity_threshold : float
        Sequences with at least this pairwise identity will be
        grouped in the same cluster.

    Returns
    -------
    np.array
        Vector of length N containing number of cluster
        members for each sequence (inverse of sequence
        weight)
    """
    N, L = matrix.shape
    L = 1.0 * L

    # minimal cluster size is 1 (self)
    num_neighbors = np.ones((N))

    # compare all pairs of sequences
    for i in range(N - 1):
        for j in range(i + 1, N):
            pair_id = 0
            for k in range(L):
                if matrix[i, k] == matrix[j, k]:
                    pair_id += 1

            if pair_id / L >= identity_threshold:
                num_neighbors[i] += 1
                num_neighbors[j] += 1

    return num_neighbors



import os
import unittest
from unittest import TestCase
from evcouplings.utils import SubmitterFactory, Command
from evcouplings.utils.app import app


class TestSubmitter(TestCase):

    def test_submitter_factory_available(self):
        print(SubmitterFactory.available_methods())

    def test_submitter_factory_init(self):
        lsf = SubmitterFactory("lsf", blocking=True, db_path="test.db")
        assert lsf.isBlocking == True

    # def minimal_LSF_example(self):
    #     lsf = SubmitterFactory("lsf")
    #     c = Command("sleep 1h",
    #                 name="test",
    #                 environment="source /home/bs224/.bashrc",
    #                 workdir="/home/bs224/",
    #                 resources={EResource.time: "01:01", EResource.queue: "short"})
    #
    #     job_id = lsf.submit(c)
    #     print(lsf.monitor(c))
    #     lsf.cancle(c)
    #     print(lsf.mointor(c))

    def test_local_submitter(self):
        local = SubmitterFactory("local", blocking=True, db_path="./test/test.db")
        c = Command("sleep 5 && touch ./test/test_submission.txt", name="test_sleep")

        local.submit(c)
        assert local.monitor(c) in ["done", "susp", "run", "pend"]
        local.join()
        assert local.monitor(c) in ["done", "exit"]
        assert os.path.exists("./test/test_submission.txt")
        os.remove("./test/test_submission.txt")
        os.remove("./test/test.db")

    def test_local_cancel(self):
        local = SubmitterFactory("local", blocking=True, db_path="./test/test.db")
        c = Command("sleep 5 && touch ./test/test_submission.txt", name="test_sleep")

        local.submit(c)
        assert local.cancel(c)
        os.remove("./test/test.db")

    def test_local_dependency(self):
        local = SubmitterFactory("local", blocking=True, db_path="./test/test.db")
        c1 = Command("sleep 5 ", name="test_sleep")
        c2 = Command("touch ./test/test_submission.txt", name="test_touch")

        local.submit(c1)
        local.submit(c2, dependent=c1)
        x = local.monitor(c2)

        assert  x in ["susp", "pend"]
        local.join()
        assert os.path.exists("./test/test_submission.txt")
        os.remove("./test/test_submission.txt")
        os.remove("./test/test.db")


if __name__ == '__main__':
    unittest.main()


"""
Visualization of evolutionary couplings (contact maps etc.)

Authors:
  Thomas A. Hopf
"""

from copy import deepcopy
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from evcouplings.visualize.pymol import (
    pymol_pair_lines, pymol_mapping
)

# default plotting styles
STYLE_EC = {
    "edgecolor": "none",
    "c": "black",
    "s": 80,
}

STYLE_CONTACT = {
    "edgecolor": "none",
    "c": "#b6d4e9",
    "s": 100,
}

STYLE_CONTACT_BRIGHT = {
    "edgecolor": "none",
    "c": "#d9e7f5",
    "s": 100,
}

STYLE_CONTACT_MULTIMER = {
    "edgecolor": "none",
    "c": "#fc8c3b",
    "alpha": 0.3,
    "s": 100,
}

STYLE_SECSTRUCT = {
    "helix_turn_length": 2,
    "strand_width_factor": 0.5,
    "min_sse_length": 2,
    "width": 1,
    "line_width": 2,
    "helix_color": "k",
    "strand_color": "k",
    "coil_color": "k",
}


def find_boundaries(boundaries, ecs, monomer, multimer, symmetric):
    """
    Identify axis boundaries for contact map plots

    Parameters
    ----------
    boundaries : {"union", "intersection", "ecs", "structure"} or tuple
             or list(tuple, tuple)
        Set axis range (min/max) of contact map as follows:

        * "union": Positions either in ECs or 3D structure
        * "intersection": Positions both in ECs and 3D structure
        * "ecs": Positions in ECs
        * "structure": Positions in 3D structure
        * tuple(float, float): Specify upper/lower bound manually
        * [(float, float), (float, float)]: Specify upper/lower bounds
          for both x-axis (first tuple) and y-axis (second tuple)

    ecs : pandas.DataFrame
        Table of evolutionary couplings to plot (using columns
        "i" and "j")
    monomer : evcouplings.compare.distances.DistanceMap
        Monomer distance map (intra-chain distances)
    multimer : evcouplings.compare.distances.DistanceMap
        Multimer distance map (multimeric inter-chain distances)
    symmetric : bool
        Sets if distance maps and ECs are symmetric (intra-chain or homomultimer),
        or not (inter-chain).

    Returns
    -------
    (min_x, max_x) : (float, float)
        First and last position on x-axis
    (min_y, max_y) : (float, float)
        First and last position on y-axis
    """
    def _find_pos(axis):
        """
        Find first and last index along a single contact map axis
        """
        # determine what sets of positions are for ECs/contact maps
        ec_pos = set()
        monomer_pos = set()
        multimer_pos = set()

        # need to merge i and j here if symmetric
        if ecs is not None:
            if symmetric:
                ec_pos = set(ecs.i.astype(int)).union(ecs.j.astype(int))
            else:
                ec_pos = set(getattr(ecs, axis).astype(int))

        if monomer is not None:
            monomer_pos = set(
                getattr(monomer, "residues_" + axis).id.astype(int)
            )

        if multimer is not None:
            multimer_pos = set(
                getattr(multimer, "residues_" + axis).id.astype(int)
            )

        structure_pos = monomer_pos.union(multimer_pos)

        # maximum ranges spanned by structure or ECs
        # if any of the sets is not given, revert to
        # the other set of positions in else case
        # (in these cases, union and intersection will
        # be trivially the one set that is actually defined)
        if len(ec_pos) > 0:
            min_ec, max_ec = min(ec_pos), max(ec_pos)
        else:
            min_ec, max_ec = min(structure_pos), max(structure_pos)

        if len(structure_pos) > 0:
            min_struct, max_struct = min(structure_pos), max(structure_pos)
        else:
            min_struct, max_struct = min(ec_pos), max(ec_pos)

        # determine and set plot boundaries
        if boundaries == "union":
            min_val = min(min_ec, min_struct)
            max_val = max(max_ec, max_struct)
        elif boundaries == "intersection":
            min_val = max(min_ec, min_struct)
            max_val = min(max_ec, max_struct)
        elif boundaries == "ecs":
            min_val = min_ec
            max_val = max_ec
        elif boundaries == "structure":
            min_val = min_struct
            max_val = max_struct
        else:
            raise ValueError(
                "Not a valid value for boundaries: {}".format(
                    boundaries
                )
            )

        return min_val, max_val

    # check first if range is specified manually
    if isinstance(boundaries, tuple):
        if len(boundaries) != 2:
            raise ValueError(
                "boundaries must be a tuple with 2 elements (min, max)."
            )
        min_x, max_x = boundaries
        min_y, max_y = boundaries
    elif isinstance(boundaries, list):
        if len(boundaries) != 2 or len(boundaries[0]) != 2 or len(boundaries[1]) != 2:
            raise ValueError(
                "boundaries must be a list of 2 tuples with 2 elements "
                "[(min_x, max_x), (min_y, max_y)]."
            )
        min_x, max_x = boundaries[0]
        min_y, max_y = boundaries[1]
    else:
        min_x, max_x = _find_pos("i")
        min_y, max_y = _find_pos("j")

    return (min_x, max_x), (min_y, max_y)


def plot_contact_map(ecs=None, monomer=None, multimer=None,
                     distance_cutoff=5, secondary_structure=None,
                     show_secstruct=True, scale_sizes=True,
                     ec_style=STYLE_EC, monomer_style=STYLE_CONTACT,
                     multimer_style=STYLE_CONTACT_MULTIMER,
                     secstruct_style=STYLE_SECSTRUCT,
                     margin=5, invert_y=True, boundaries="union",
                     symmetric=True, ax=None):
    """
    Wrapper for simple contact map plots with optional
    multimer contacts (can also handle non-symmetric inter-chain
    contacts). For full flexibility, compose your own
    contact map plot using the functions used below.

    Parameters
    ----------
    ecs : pandas.DataFrame
        Table of evolutionary couplings to plot (using columns
        "i" and "j"). Can contain additional columns "color"
        and "size" to assign these individual properties in the
        plot for each plotted pair (i, j). If all values of
        "size" are <= 1, they will be treated as a fraction
        of the point size defined in ec_style, and rescaled
        if scale_sizes is True.
    monomer : evcouplings.compare.distances.DistanceMap
        Monomer distance map (intra-chain distances)
    multimer : evcouplings.compare.distances.DistanceMap
        Multimer distance map (multimeric inter-chain distances)
    distance_cutoff : float, optional (default: 5)
        Pairs with distance <= this cutoff are considered
        residue pair contacts
    secondary_structure : dict or pandas.DataFrame
        Secondary structure to be displayed on both axis
        (if not given, will try to extract from monomer
        distance map). For format of dict or DataFrame,
        see documentation of plot_secondary_structure().
        If symmetric is False, this has to be a two-element
        tuple containing the secondary structures for the
        x-axis and the y-axis, respectively.
    show_secstruct : bool, optional (default: True)
        Draw secondary structure on both axes (either
        passed in explicitly using secondary_structure,
        or extracted from monomer distancemap)
    scale_sizes : bool, optional (default: True)
        Rescale sizes of points on scatter plots as well as
        secondery structure plotting width based on
        overall size of protein
    ec_style : dict, optional (default: STYLE_EC)
        Style for EC plotting (kwargs to plt.scatter)
    monomer_style : dict, optional (default: STYLE_CONTACT)
        Style for monomer contact plotting (kwargs to plt.scatter)
    multimer_style : dict, optional (default: STYLE_CONTACT_MULTIMER)
        Style for multimer contact plotting (kwargs to plt.scatter)
    secstruct_style : dict, optional (default: STYLE_SECSTRUCT)
        Style for secondary structure plotting
        (kwargs to secondary_structure_cartoon())
    margin : int, optional (default: 5)
        Space to add around contact map
    invert_y : bool, optional (default: True)
        Invert the y axis of the contact map so both sequences
        run from N -to C- terminus starting from top left corner
    boundaries : {"union", "intersection", "ecs", "structure"} or tuple or list(tuple, tuple), optional (default: "union")
        Set axis range (min/max) of contact map as follows:

        * "union": Positions either in ECs or 3D structure
        * "intersection": Positions both in ECs and 3D structure
        * "ecs": Positions in ECs
        * "structure": Positions in 3D structure
        * tuple(float, float): Specify upper/lower bound manually
        * [(float, float), (float, float)]: Specify upper/lower bounds
          for both x-axis (first tuple) and y-axis (second tuple)

    symmetric : bool, optional (default: True)
        Sets if distance maps and ECs are symmetric (intra-chain or homomultimer),
        or not (inter-chain).
    ax : Matplotlib Axes object, optional (default: None)
        Axes the plot will be drawn on
    """
    if ecs is None and monomer is None and multimer is None:
        raise ValueError(
            "Need to specify at least one of ecs, monomer or multimer"
        )

    if ax is None:
        ax = plt.gca()

    # figure out how to set contact map boundaries
    (min_x, max_x), (min_y, max_y) = find_boundaries(
        boundaries, ecs, monomer, multimer, symmetric
    )

    set_range(
        x=(min_x, max_x), y=(min_y, max_y),
        ax=ax, margin=margin, invert_y=invert_y
    )

    # enable rescaling of points and secondary structure if necessary
    if scale_sizes:
        scale_func = lambda x: scale(x, ax=ax)
    else:        
        scale_func = lambda x: x

    # plot monomer contacts
    # (distance maps will automatically be symmetric for
    # intra/homomultimer, so do not request mirroring)
    if monomer is not None:
        plot_pairs(
            monomer.contacts(distance_cutoff),
            symmetric=False, style=scale_func(monomer_style),
            ax=ax
        )

    # plot multimer contacts
    # (distance maps will automatically be symmetric for
    # intra/homomultimer, so again do not request mirroring)
    if multimer is not None:
        plot_pairs(
            multimer.contacts(distance_cutoff),
            symmetric=False, style=scale_func(multimer_style),
            ax=ax
        )

    # plot ECs
    # (may be symmetric or not, depending on use case)
    if ecs is not None:
        plot_pairs(
            ecs, symmetric=symmetric, style=scale_func(ec_style), ax=ax
        )

    # plot secondary structure
    if show_secstruct:
        # if secondary structure given explicitly, use it
        if secondary_structure is not None:
            if symmetric:
                secstruct_i = secondary_structure
                secstruct_j = secondary_structure
            else:
                if (not isinstance(secondary_structure, tuple) or
                        len(secondary_structure) != 2):
                    raise ValueError(
                        "When symmetric is False, secondary structure must "
                        "be a tuple (secstruct_i, secstruct_j)."
                    )
                secstruct_i, secstruct_j = secondary_structure

            plot_secondary_structure(
                secstruct_i, secstruct_j,
                style=scale_func(secstruct_style), ax=ax
            )
        else:
            # otherwise, see if we can extract it from monomer
            # distance map
            try:
                plot_secondary_structure(
                    monomer.residues_i, monomer.residues_j,
                    style=scale_func(secstruct_style), ax=ax
                )
            except AttributeError:
                # DataFrame has no secondary structure, cannot do anything here
                # (this happens when merging across multiple distance maps)
                pass


def plot_pairs(pairs, symmetric=False, ax=None, style=None):
    """
    Plot list of pairs (ECs/contacts)

    Parameters
    ----------
    pairs : pandas.DataFrame
        DataFrame with coordinates to plot
        (taken from columns i and j). If there
        are columns "color" and "size", these
        will be used to assign individual colors
        and sizes to the dots in the scatter plot.
        If sizes are all <= 1 and "s" is present as a
        key in style, values will be treated as fraction
        of "s".
    symmetric : bool, optional (default: False)
        If true, for each pair (i, j) also plot
        pair (j, i). This is for cases where
        the input list pairs does not contain
        both pairs.
    ax : matplotlib Axes object
        Axes to plot on
    style : dict
        Parameters to style pair scatter plot
        (passed as **kwargs to matplotlib plt.scatter())

    Returns
    -------
    paths : list of PathCollection
        Scatter plot paths drawn by this function
    """
    if ax is None:
        ax = plt.gca()

    ax.set_aspect("equal")

    if style is None:
        style = {}

    if "color" in pairs.columns:
        style["c"] = pairs.loc[:, "color"].values

    if "size" in pairs.columns:
        # if all sizes <= 1, treat as fraction
        if len(pairs.query("size > 1")) == 0 and "s" in style:
            style["s"] *= pairs.loc[:, "size"].values
        # otherwise take as actual value
        else:
            style["s"] = pairs.loc[:, "size"].values

    path1 = ax.scatter(
        pairs.i.astype(int),
        pairs.j.astype(int),
        **style
    )

    paths = [path1]

    if symmetric:
        path2 = ax.scatter(
            pairs.j.astype(int),
            pairs.i.astype(int),
            **style
        )
        paths.append(path2)

    return paths


def set_range(pairs=None, symmetric=True, x=None, y=None,
              ax=None, margin=0, invert_y=True):
    """
    Set axes ranges for contact map based
    on minimal/maximal values in pair list

    Parameters
    ----------
    pairs : pandas.DataFrame, optional (default: None)
        DataFrame with pairs (will be extracted
        from columns named "i" and "j" and
        converted to integer values and used
        to define x- and y-axes, respectively).
        If None, if x and y have to be specified.
    symmetric : bool, optional (default:True)
        If true, will define range on joint set
        of values in columns i and j, resulting
        in a square contact map
    x : tuple(float, float), optional (default: None)
        Set x-axis range with this range (min, max).
        Will be extended by margin, and overrides any
        value for x-axis derived using pairs.
    y : tuple(float, float), optional (default: None)
        Set y-axis range with this range (min, max)
        Will be extended by margin, and overrides any
        value for y-axis derived using pairs.
    ax : matplotlib Axes object
        Axes for which plot range will be changed
    margin : int, optional (default: 0)
        Add margin of this size around the actual
        range defined by the data
    invert_y : bool, optional (default: True)
        Invert y-axis of contact map

    Returns
    -------
    x_range : tuple(int, int)
        Set range for x-axis
    y_range : tuple(int, int)
        Set range for y-axis

    Raises
    ------
    ValueError
        If any axis range remains unspecified
    """
    if ax is None:
        ax = plt.gca()

    x_range, y_range = None, None

    # infer plot range from data
    if pairs is not None:
        i = pairs.i.astype(int)
        j = pairs.j.astype(int)

        if symmetric:
            x_range = (
                min(i.min(), j.min()) - margin,
                max(i.max(), j.max()) + margin
            )
            y_range = x_range
        else:
            x_range = (i.min() - margin, i.max() + margin)
            y_range = (j.min() - margin, j.max() + margin)

    # Override with user-specified values
    if x is not None:
        x_range = (x[0] - margin, x[1] + margin)

    if y is not None:
        y_range = (y[0] - margin, y[1] + margin)

    if x_range is None or y_range is None:
        raise ValueError(
            "Axis remained unspecified (make sure to either "
            "set pairs pr x_range/y_range) :"
            " x: {} y: {}".format(
                x_range, y_range
            )
        )

    # maintain axis inversion
    # (which gets undone by setting x/ylim)
    inverted_x = ax.xaxis_inverted()
    inverted_y = ax.yaxis_inverted()

    # set new range
    ax.set_xlim(x_range)
    ax.set_ylim(y_range)

    # recover axis orientation
    if inverted_x:
        ax.invert_xaxis()

    if inverted_y or (invert_y and not inverted_y):
        ax.invert_yaxis()

    # make sure axis ticks are in right spot
    ax.yaxis.set_ticks_position("left")
    if ax.yaxis_inverted():
        ax.xaxis.set_ticks_position("top")
    else:
        ax.xaxis.set_ticks_position("bottom")

    return x_range, y_range


def scale(style, ax=None):
    """
    Scale size of drawn elements based on size
    of contact map plot
    """
    if ax is None:
        ax = plt.gca()

    x_range = ax.get_xlim()
    y_range = ax.get_ylim()

    L = max(
        abs(x_range[1] - x_range[0]),
        abs(y_range[1] - y_range[0])
    )
    style = deepcopy(style)

    # dot size
    if "s" in style.keys():
        style["s"] = style["s"]**2 / L

    # secondary structure width
    if "width" in style.keys():
        style["width"] = style["width"] * L / 100

    return style


def plot_secondary_structure(secstruct_i, secstruct_j=None, ax=None, style=None, margin=None):
    """
    Plot secondary structure along contact map.

    Note: this function should only be used *after* the
    orientation of the axes of the plot has been set.

    Parameters
    ----------
    secstruct_i : dict or pd.DataFrame
        Secondary structure for x-axis of plot.
        Can be a dictionary of position (int) to
        secondary structure character ("H", "E", "C", "-"),
        or a DataFrame with columns "id" and "sec_struct_3state"
        (as returned by Chain.residues, and DistanceMap.residues_i
        and DistanceMap.residues_j).
    secstruct_j : dict or pd.DataFrame, optional (default: None)
        Secondary structure for y-axis of plot.
        See secstruct_i for possible values.
        If None, will use secstruct_i for y-axis too
        (assuming symmetric contact map).
    ax : matplotlib Axes object
        Axis to draw secondary structure on
    style : dict, optional (default: None)
        Parameters that will be passed on as kwargs
        to secondary structure drawing function
    margin : int, optional (default: None)
        Add this much space between contact map
        and secondary structure. If None, defaults
        to the width of secondary structure * 3.
    """
    def _extract_secstruct(secstruct, axis_range):
        # turn into dictionary representation if
        # passed as a DataFrame
        if isinstance(secstruct, pd.DataFrame):
            # first check we actually have a secondary
            # structure column
            if "sec_struct_3state" not in secstruct.columns:
                return None, None, None

            # do not store any undefined secondary
            # structure in dictionary, or NaN
            # values will lead to problems

            secstruct = secstruct.dropna(
                subset=["sec_struct_3state"]
            )

            secstruct = dict(
                zip(
                    secstruct.id.astype(int),
                    secstruct.sec_struct_3state
                )
            )

        # catch case where there is no secondary
        # structure at all (e.g. because of dataframe
        # full of NaNs)
        if len(secstruct) == 0:
            return None, None, None

        # make sure we only retain secondary structure
        # inside the range of the plot, otherwise
        # drawing artifacts occur
        range_min = min(axis_range)
        range_max = max(axis_range)
        secstruct = {
            i: sstr for (i, sstr) in secstruct.items()
            if range_min <= i < range_max
        }

        first_pos, last_pos = min(secstruct), max(secstruct) + 1
        secstruct_str = "".join(
            [secstruct.get(i, "-") for i in range(first_pos, last_pos)]
        )

        start, end, segments = find_secondary_structure_segments(
            secstruct_str, offset=first_pos
        )

        return start, end, segments

    if ax is None:
        ax = plt.gca()

    if style is None:
        style = {}

    # make secondary structure symmetric if not given
    if secstruct_j is None:
        secstruct_j = secstruct_i

    # get axis ranges to place secondary structure drawings
    x_range = ax.get_xlim()
    y_range = ax.get_ylim()

    # i corresponds to x-axis, j to y-axis
    if margin is None:
        margin = 3 * style.get("width", 1)
    else:
        margin += style.get("width", 1)

    start_i, end_i, segments_i = _extract_secstruct(secstruct_i, x_range)
    if segments_i is not None:
        secondary_structure_cartoon(
            segments_i,
            **{
                **style,
                "center": max(y_range) + margin,
                "ax": ax,
                "sequence_start": start_i,
                "sequence_end": end_i,
                "horizontal": True,
            }
        )

    start_j, end_j, segments_j = _extract_secstruct(secstruct_j, y_range)
    if segments_j is not None:
        secondary_structure_cartoon(
            segments_j,
            **{
                **style,
                "center": max(x_range) + margin,
                "ax": ax,
                "sequence_start": start_j,
                "sequence_end": end_j,
                "horizontal": False,
            }
        )


def secondary_structure_cartoon(
        sse, ax=None, sequence_start=0,
        sequence_end=None, horizontal=True,
        flip_direction=False, center=0, width=1,
        helix_turn_length=1, strand_width_factor=0.5,
        line_width=2, min_sse_length=0, clipping=False,
        helix_color="k", strand_color="k", coil_color="k",
        draw_coils=True
):
    """
    Plot a 1D secondary structure cartoon.

    Parameters
    ----------
    sse : list
        Secondary structure elements, as returned by
        find_secondary_structure_segments(). Tuples
        in the list have the form (type, start, end),
        where type is either H (helix), E (sheet),
        or C (coil/other)
    ax: matplotlib Axes object
        Axis to draw cartoon on
    sequence_start : float, optional (default:0)
        Plot coordinate of first position from which
        cartoon will be drawn
    sequence_end : float, optional (default: None)
        Last coordinate up to which cartoon will
        be drawn
    horizontal : bool, optional (default: True)
        If True, draw cartoon horizontally, vertically
        otherwise
    flip_direction : bool, optional (default: False)
        Invert drawing direction (from right to left)
    center : float, optional (default: True)
        Center plot coordinate along which cartoon
        will be drawn
    width : float, optional (default: 1)
        Width of secondary structure cartoon
    helix_turn_length : float, optional (default: 1)
        Length for a full helix turn in plot coordinates
        (number of residues per sine)
    strand_width_factor : float, optional (default: 0.5)
        Width of strand relative to full width of cartoon
    line_width : float, optional (default: 2)
        Line width for drawing
    min_sse_length : float, optional (default: 0)
        Only draw secondary structure elements with
        length greater than this threshold
    clipping : bool, optional (default: False)
        Clip drawing at plot axis (must be False
        to draw outside contact map area)
    helix_color : str or tuple, optional (default: "k")
        Matplotlib color to be used for helices
    strand_color : str or tuple, optional (default: "k")
        Matplotlib color to be used for beta strands
    coil_color : str or tuple, optional (default: "k")
        Matplotlib color to be used for all other positions
    draw_coils : bool, optional (Default: True)
        If true, draw line for coil segments.
    """
    def _transform(x, y):
        """
        Transform raw drawing coordinates if
        axis or direction is flipped
        """
        x, y = np.array(x), np.array(y)
        x = -x if flip_direction else x
        return (x, y) if horizontal else (y, x)

    if ax is None:
        ax = plt.gca()

    # collect segments that are neither helix nor
    # strand - this is important since ends of
    # helix segments in plot are not exactly known
    # a priori
    no_ss_segments = [sequence_start]

    # draw secondary structure segments one by one
    for (ss_type, start, end) in sse:
        # do not draw very short elements if chosen
        if end - start < min_sse_length:
            continue

        if ss_type == "H":  # alpha helix
            # length of segment (do not add 1
            # since the number of intervals is
            # one less than the number of residues)
            length = end - start

            # unrounded number of helix turns
            turns = length / float(helix_turn_length)

            # round up to number of complete turns
            # (otherwise helix ends not aligned);
            # make sure there is at least one turn,
            # or code will crash
            full_turns = max(1.0, np.ceil(turns))

            # mismatch between residues covered by rounded
            # number of turns and how many turns the segments
            # actually covers
            overhang = full_turns - turns

            x_sin = np.arange(round(full_turns), step=0.02)
            y = np.sin(2 * np.pi * x_sin) * width + center
            x = start + (x_sin - overhang / 2.0) * helix_turn_length
            ax.plot(
                *_transform(x, y), color=helix_color,
                ls="-", lw=line_width, clip_on=clipping
            )

            # store beginning and end coordinates
            # of helix to fill coil in between

            no_ss_segments.append(x.min())
            no_ss_segments.append(x.max())

        elif ss_type == "E":  # beta strand
            # rectangle part
            if abs(end - start) > 1:
                x_t, y_t = _transform(
                    [start, end - width],
                    [center - width * strand_width_factor,
                     center + width * strand_width_factor]
                )
                ax.add_patch(
                    plt.Rectangle(
                        (x_t[0], y_t[0]),
                        x_t[1] - x_t[0], y_t[1] - y_t[0],
                        edgecolor=strand_color,
                        facecolor=strand_color,
                        clip_on=clipping
                    )
                )

            # triangle part
            x = [end - width, end - width, end]
            y = [center - width, center + width, center]
            ax.add_patch(
                plt.Polygon(
                    list(zip(*_transform(x, y))),
                    edgecolor=strand_color, facecolor=strand_color,
                    clip_on=clipping
                )
            )

            # store beginning and end coordinates
            # of strand to fill coil in between
            no_ss_segments.append(start)
            no_ss_segments.append(end)

        elif ss_type == "-":  # skip drawing (no data)
            no_ss_segments.append(start)
            no_ss_segments.append(end)

    # draw coil until given endpoint
    if sequence_end is not None:
        no_ss_segments.append(sequence_end + 1)

    # finally, draw all coil segments
    if draw_coils:
        for (start, end) in zip(
            no_ss_segments[::2],
            no_ss_segments[1::2]
        ):
            if start > end:
                continue

            x = [start, end]
            y = [center, center]
            ax.plot(*_transform(x, y), color=coil_color,
                    ls="-", lw=line_width, clip_on=clipping)


def find_secondary_structure_segments(sse_string, offset=0):
    """
    Identify segments of secondary structure elements in string

    Parameters
    ----------
    sse_string : str
        String with secondary structure states of sequence
        ("H", "E", "-"/"C")
    offset : int, optional (default: 0)
        Shift start/end indices of segments by this offset

    Returns
    -------
    start : int
        Index of first position (equal to "offset")
    end : int
        Index of last position
    segments : list
        List of tuples with the following elements:

        1. secondary structure element (str)
        2. start position of segment (int)
        3. end position of segment, exlusive (int)
    """
    if len(sse_string) < 1:
        raise ValueError("Secondary structure string must have length > 0.")

    end = len(sse_string) - 1

    sse_list = list(sse_string)
    change_points = [
        (i, (c1, c2)) for (i, (c1, c2)) in
        enumerate(zip(sse_list[:-1], sse_list[1:]))
        if c1 != c2
    ]

    segments = []
    last_start = 0
    # set s2 for the case of only one continuous segment
    s2 = sse_string[0]
    for (p, (s1, s2)) in change_points:
        segments.append((s1, offset + last_start, offset + p + 1))
        last_start = p + 1
    segments.append((s2, offset + last_start, offset + end + 1))

    return offset, end + offset, segments


def ec_lines_pymol_script(ec_table, output_file, distance_cutoff=5,
                          score_column="cn", chain=None):
    """
    Create a Pymol .pml script to visualize ECs on a 3D
    structure
    
    Parameters
    ----------
    ec_table : pandas.DataFrame
        Visualize all EC pairs (columns i, j) in this
        table. If a column "dist" exists and distance_cutoff
        is defined, will assign different colors based on
        the 3D distance of the EC.
    output_file : str
        File path where to store pml script
    distance_cutoff : float, optional (default: 5)
        Color ECs with distance above this threshold
        as false positives (only possible if a column
        "dist" exists in ec_table). If None, will
        use one color for all ECs.
    score_column : str, optional (default: "cn")
        Use this column in ec_table to adjust radius
        of lines. If None, all lines will be drawn
        at equal radius.
    chain : str, optional (default: None)
        Use this PDB chain in residue selection
    """
    t = ec_table.copy()

    # assign line styles
    for prop, val in [
        ("dash_radius", 0.345), ("dash_gap", 0.075), ("dash_length", 0.925)
    ]:
        t.loc[:, prop] = val

    # adjust line width/radius based on score, if selected
    if score_column is not None:
        scaling_factor = 0.5 / ec_table.loc[:, score_column].max()
        t.loc[:, "dash_radius"] = ec_table.loc[:, score_column] * scaling_factor

    if "dist" in ec_table and distance_cutoff is not None:
        t.loc[t.dist <= distance_cutoff, "color"] = "green"
        t.loc[t.dist > distance_cutoff, "color"] = "red"
    else:
        t.loc[:, "color"] = "green"

    if chain is not None:
        chain_sel = ", chain '{}'".format(chain)
    else:
        chain_sel = ""

    with open(output_file, "w") as f:
        f.write("as cartoon{}\n".format(chain_sel))
        f.write("color grey80{}\n".format(chain_sel))
        pymol_pair_lines(t, f, chain)


def enrichment_pymol_script(enrichment_table, output_file,
                            sphere_view=True, chain=None):
    """
    Create a Pymol .pml script to visualize EC "enrichment"

    Parameters
    ----------
    enrichment_table : pandas.DataFrame
        Mapping of position (column i) to EC enrichment
        (column enrichemnt), as returned by 
        evcouplings.couplings.pairs.enrichment()
    output_file : str
        File path where to store pml script
    sphere_view : bool, optional (default: True)
        If True, create pml that highlights enriched positions
        with spheres and color; if False, create pml
        that highlights enrichment using b-factor and
        "cartoon putty"
    chain : str, optional (default: None)
        Use this PDB chain in residue selection
    """
    t = enrichment_table.query("enrichment > 1")

    # compute boundaries for highly coupled residues
    # that will be specially highlighted
    boundary1 = int(0.05 * len(t))  # top 5%
    boundary2 = int(0.15 * len(t))  # top 15%

    t.loc[:, "b_factor"] = t.enrichment

    # set color for "low" enrichment (anything > 1)
    t.loc[:, "color"] = "yelloworange"

    # high
    t.loc[t.iloc[0:boundary1].index, "color"] = "red"

    # medium
    t.loc[t.iloc[boundary1:boundary2].index, "color"] = "orange"

    if sphere_view:
        t.loc[t.iloc[0:boundary2].index, "show"] = "spheres"

    if chain is not None:
        chain_sel = ", chain '{}'".format(chain)
    else:
        chain_sel = ""

    with open(output_file, "w") as f:
        f.write("as cartoon{}\n".format(chain_sel))
        f.write("color grey80{}\n".format(chain_sel))

        if chain is None:
            f.write("alter all, b=0.0\n")
        else:
            f.write("alter chain '{}', b=0.0\n".format(chain))

        pymol_mapping(t, f, chain)

        if not sphere_view:
            f.write("cartoon putty{}\n".format(chain_sel))



"""
Useful values and constants for all of package

Authors:
  Thomas A. Hopf
"""

# amino acid one-letter code to three-letter code
AA1_to_AA3 = {
    "A": "ALA",
    "B": "ASX",
    "C": "CYS",
    "D": "ASP",
    "E": "GLU",
    "F": "PHE",
    "G": "GLY",
    "H": "HIS",
    "I": "ILE",
    "K": "LYS",
    "L": "LEU",
    "M": "MET",
    "N": "ASN",
    "P": "PRO",
    "Q": "GLN",
    "R": "ARG",
    "S": "SER",
    "T": "THR",
    "V": "VAL",
    "W": "TRP",
    "X": "XAA",
    "Y": "TYR",
    "Z": "GLX",
}

# amino acid three-letter code to one-letter code
AA3_to_AA1 = {
    v: k for k, v in AA1_to_AA3.items()
}



from evcouplings.utils.calculations import *
from evcouplings.utils.config import *
from evcouplings.utils.helpers import *
from evcouplings.utils.system import *
from evcouplings.utils.batch import *
from evcouplings.utils.constants import *

class ASubmitterFactory(abc.ABCMeta):
    def __init__(cls, name, bases, nmspc):
        type.__init__(cls, name, bases, nmspc)

    def __call__(self, _name, *args, **kwargs):
        '''
        If a third person wants to write a new Submitter. He/She has to inherit from ASubmitter.
        That's it nothing more.
        '''

        try:
            return ASubmitter[str(_name).lower()](**kwargs)
        except KeyError as e:
            raise ValueError("This submitter is currently not supported")


class SubmitterFactory(object, metaclass=ASubmitterFactory):


    @staticmethod
    def available_methods():
        """
        Returns a dictionary of available epitope predictors and their supported versions
        :return: dict(str,list(str) - dictionary of epitope predictors represented as string and a list of supported
                                      versions
        """
        return [ASubmitter.registry.keys()]




"""
Mapping indices for complexes / multi-domain sequences to
internal model numbering.

Authors:
  Thomas A. Hopf
"""

from collections import Iterable
from copy import deepcopy
import pandas as pd


class Segment:
    """
    Represents a continuous stretch of sequence in a sequence
    alignment to infer evolutionary couplings (e.g. multiple domains,
    or monomers in a concatenated complex alignment)
    """
    def __init__(self, segment_type, sequence_id,
                 region_start, region_end, positions=None,
                 segment_id="A"):
        """
        Create a new sequence segment

        Parameters
        ----------
        segment_type : {"aa", "dna", "rna"}
            Type of sequence
        sequence_id : str
            Identifier of sequence
        region_start : int
            Start index of sequence segment
        region_end : int
            End index of sequence segment (position
            is inclusive)
        positions : list(int), optional (default: None)
            Positions in the sequence alignment that
            will be used for EC calculation
            (all positions corresponding to uppercase
            residues). Compulsory parameter when using
            non-focus mode.
        segment_id : str
            Identifier of segment (must be unique)
        """
        self.segment_type = segment_type
        self.sequence_id = sequence_id
        self.region_start = region_start
        self.region_end = region_end
        if positions is not None:
            self.positions = list(map(int, positions))
        else:
            self.positions = None

        self.segment_id = segment_id

    @classmethod
    def from_list(cls, segment):
        """
        Create a segment object from list representation
        (e.g. from config).

        Parameters
        ----------
        segment : list
            List representation of segment, with the following items:
            segment_id (str), segment_type (str), sequence_id (str),
            region_start (int), region_end (int), positions (list(int))

        Returns
        -------
        Segment
            New Segment instance from list
        """
        segment_id, segment_type, sequence_id, region_start, region_end, positions = segment
        return cls(segment_type, sequence_id, region_start, region_end, positions, segment_id)

    def to_list(self):
        """
        Represent segment as list (for storing in configs)

        Returns
        -------
        list
            List representation of segment, with the following items:
            segment_id (str), segment_type (str), sequence_id (str),
            region_start (int), region_end (int), positions (list(int))
        """
        return [
            self.segment_id,
            self.segment_type,
            self.sequence_id,
            self.region_start,
            self.region_end,
            self.positions
        ]


class SegmentIndexMapper:
    """
    Map indices of one or more sequence segments into
    CouplingsModel internal numbering space. Can also
    be used to (trivially) remap indices for a single sequence.
    """
    def __init__(self, focus_mode, first_index, *segments):
        """
        Create index mapping from individual segments

        Parameters
        ----------
        focus_mode : bool
            Set to true if model was inferred in focus mode,
            False otherwise.
        first_index : int
            Index of first position in model/sequence.
            For nonfocus mode, should always be one. For focus
            mode, corresponds to index given in sequence header
            (1 if not in alignment)
        *segments: (int, int):
            Segments containing numberings for each
            individual segment
        """
        # store segments so we retain full information
        self.segments = deepcopy(segments)

        # build up target indices by going through all segments
        self.target_pos = []
        for s in segments:
            if focus_mode:
                # in focus mode, we simply assemble the
                # ranges of continuous indices, because
                # numbering in model is also continuous
                cur_target = range(
                    s.region_start, s.region_end + 1
                )
            else:
                # in non-focus mode, we need to assemble
                # the indices of actual model positions,
                # since the numbering in model may be
                # discontinuous
                cur_target = s.positions

            # create tuples of (segment_id, target_pos)
            self.target_pos += list(zip(
                [s.segment_id] * len(cur_target),
                cur_target
            ))

        # create correspond list of model positions;
        # note that in focus mode, not all of these
        # positions might actually be in the model
        # (if they correspond to lowercase columns)
        self.model_pos = list(range(
            first_index, first_index + len(self.target_pos)
        ))

        # mapping from target sequences (segments) into
        # model numbering space (continuous numbering)
        self.target_to_model = dict(
            zip(self.target_pos, self.model_pos)
        )

        # inverse mapping from model numbering into target
        # numbering
        self.model_to_target = dict(
            zip(self.model_pos, self.target_pos)
        )

    def patch_model(self, model, inplace=True):
        """
        Change numbering of CouplingModel object
        so that it uses segment-based numbering

        Parameters
        ----------
        model : CouplingsModel
            Model that will be updated to segment-
            based numbering
        inplace : bool, optional (default: True)
            If True, change passed model; otherwise
            returnnew object

        Returns
        -------
        CouplingsModel
            Model with updated numbering
            (if inplace is False, this will
            point to original model)

        Raises
        ------
        ValueError
            If segment mapping does not match
            internal model numbering
        """
        if not inplace:
            model = deepcopy(model)

        try:
            mapped = [
                self.model_to_target[pos]
                for pos in model.index_list
            ]
        except KeyError:
            raise ValueError(
                "Mapping from target to model positions does "
                "not contain all positions of internal model numbering"
            )

        # update model mapping
        model.index_list = mapped

        # return updated model
        return model

    def __map(self, indices, mapping_dict):
        """
        Applies index mapping either to a single index,
        or to a list of indices

        Parameters
        ----------
        indices: int, or (str, int), or lists thereof
            Indices in input numbering space

        mapping_dict : dict(int->(str, int)) or dict((str, int)-> int)
            Mapping from one indexing space into the other

        Returns
        -------
        list of int, or list of (str, int)
            Mapped indices
        """
        if isinstance(indices, Iterable) and not isinstance(indices, tuple):
            return [mapping_dict[x] for x in indices]
        else:
            return mapping_dict[indices]

    def __call__(self, segment_id, pos):
        """
        Function-style syntax for single position to be mapped
        (calls to_model method)

        Parameters
        ----------
        segment_id : str
            Identifier of segment
        pos : int
            Position in segment numbering

        Returns
        -------
        int
            Index in coupling object numbering space
        """
        return self.to_model((segment_id, pos))

    def to_target(self, x):
        """
        Map model index to target index

        Parameters
        ----------
        x : int, or list of ints
            Indices in model numbering

        Returns
        -------
        (str, int), or list of (str, int)
            Indices mapped into target numbering.
            Tuples are (segment_id, index_in_segment)
        """
        return self.__map(x, self.model_to_target)

    def to_model(self, x):
        """
        Map target index to model index

        Parameters
        ----------
        x : (str, int), or list of (str, int)
            Indices in target indexing
            (segment_id, index_in_segment)

        Returns
        -------
        int, or list of int
            Monomer indices mapped into couplings object numbering
        """
        return self.__map(x, self.target_to_model)


def segment_map_ecs(ecs, mapper):
    """
    Map EC dataframe in model numbering into
    segment numbering

    Parameters
    ----------
    ecs : pandas.DataFrame
        EC table (with columns i and j)

    Returns
    -------
    pandas.DataFrame
        Mapped EC table (with columns i and j
        mapped, and additional columns
        segment_i and segment_j)
    """
    ecs = deepcopy(ecs)

    def _map_column(col):
        seg_col = "segment_" + col
        # create new dataframe with two columns
        # 1) mapped segment, 2) mapped position
        col_m = pd.DataFrame(
            mapper.to_target(ecs.loc[:, col]),
            columns=[seg_col, col]
        )
        # assign values instead of Series, because otherwise
        # wrong values end up in column
        ecs.loc[:, col] = col_m.loc[:, col].values
        ecs.loc[:, seg_col] = col_m.loc[:, seg_col].values

    # map both position columns (and add segment id)
    _map_column("i")
    _map_column("j")

    return ecs



"""
Useful Python helpers

Authors:
  Thomas A. Hopf, Benjamin Schubert
"""

from collections import OrderedDict
import pickle, json, csv, os, shutil
from os import path
import jinja2
import sys


class PersistentDict(dict):
    ''' Persistent dictionary with an API compatible with shelve and anydbm.

    The dict is kept in memory, so the dictionary operations run as fast as
    a regular dictionary.

    Write to disk is delayed until close or sync (similar to gdbm's fast mode).

    Input file format is automatically discovered.
    Output file format is selectable between pickle, json, and csv.
    All three serialization formats are backed by fast C implementations.

    https://code.activestate.com/recipes/576642/
    '''

    def __init__(self, filename, flag='c', mode=None, format='json', *args, **kwds):
        self.flag = flag                    # r=readonly, c=create, or n=new
        self.mode = mode                    # None or an octal triple like 0644
        self.format = format                # 'csv', 'json', or 'pickle'
        self.filename = filename
        if flag != 'n' and os.access(filename, os.R_OK):
            fileobj = open(filename, 'rb' if format=='pickle' else 'r')
            with fileobj:
                self.load(fileobj)
        dict.__init__(self, *args, **kwds)

    def sync(self):
        'Write dict to disk'
        if self.flag == 'r':
            return
        filename = self.filename
        tempname = filename + '.tmp'
        fileobj = open(tempname, 'wb' if self.format=='pickle' else 'w')
        try:
            self.dump(fileobj)
        except Exception:
            os.remove(tempname)
            raise
        finally:
            fileobj.close()
        shutil.move(tempname, self.filename)    # atomic commit
        if self.mode is not None:
            os.chmod(self.filename, self.mode)

    def close(self):
        self.sync()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def dump(self, fileobj):
        if self.format == 'csv':
            csv.writer(fileobj).writerows(self.items())
        elif self.format == 'json':
            json.dump(self, fileobj, separators=(',', ':'))
        elif self.format == 'pickle':
            pickle.dump(dict(self), fileobj, 2)
        else:
            raise NotImplementedError('Unknown format: ' + repr(self.format))

    def load(self, fileobj):
        # try formats from most restrictive to least restrictive
        for loader in (pickle.load, json.load, csv.reader):
            fileobj.seek(0)
            try:
                return self.update(loader(fileobj))
            except Exception:
                pass
        raise ValueError('File not in a supported format')


class DefaultOrderedDict(OrderedDict):
    """
    Source:
    http://stackoverflow.com/questions/36727877/inheriting-from-defaultddict-and-ordereddict
    Answer by http://stackoverflow.com/users/3555845/daniel

    Maybe this one would be better?
    http://stackoverflow.com/questions/6190331/can-i-do-an-ordered-default-dict-in-python
    """
    def __init__(self, default_factory=None, **kwargs):
        OrderedDict.__init__(self, **kwargs)
        self.default_factory = default_factory

    def __missing__(self, key):
        result = self[key] = self.default_factory()
        return result


def wrap(text, width=80):
    """
    Wraps a string at a fixed width.

    Arguments
    ---------
    text : str
        Text to be wrapped
    width : int
        Line width

    Returns
    -------
    str
        Wrapped string
    """
    return "\n".join(
        [text[i:i + width] for i in range(0, len(text), width)]
    )


def range_overlap(a, b):
    """
    Source: http://stackoverflow.com/questions/2953967/
            built-in-function-for-computing-overlap-in-python

    .. note::

        Ends of range are not inclusive

    Parameters
    ----------
    a : tuple(int, int)
        Start and end of first range
        (end of range is not inclusive)
    b : tuple(int, int)
        Start and end of second range
        (end of range is not inclusive)

    Returns
    -------
    int
        Length of overlap between ranges a and b
    """
    return max(0, min(a[1], b[1]) - max(a[0], b[0]))


def render_template(template_file, mapping):
    """
    Render a template using jinja2 and substitute
    values from mapping

    Parameters
    ----------
    template_file : str
        Path to jinja2 template
    mapping : dict
        Mapping used to substitute values
        in the template

    Returns
    -------
    str
        Rendered template
    """
    template_dir, filename = path.split(template_file)

    jinja_env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(template_dir),
        trim_blocks=True,
        lstrip_blocks=True
    )

    template = jinja_env.get_template(filename)

    return template.render(mapping)


class Progressbar(object):
    """
    Progress bar for command line programs

    Parameters
    ----------
    total_size : int
        The total size of the iteration
    bar_length : int
        The visual bar length that gets printed on stdout
    """

    def __init__(self, total_size, bar_length=60):
        self.total_size = total_size
        self.current_size = 0
        self.bar_length = bar_length


    def __iadd__(self, chunk):
        """
        Convenience function of self.update

        Parameters
        ----------
        chunk : int
            The size of the elements that are processed in the current iteration
        """
        self.update(chunk)
        return self

    def update(self, chunk):
        """
        Updates and prints the progress of the progressbar

        Parameters
        ----------
        chunk : int
            The size of the elements that are processed in the current iteration
        """

        self.current_size += chunk
        if self.current_size < self.total_size:
            filled_len = int(round(self.bar_length * self.current_size / float(self.total_size)))
            percents = round(100.0 * self.current_size / float(self.total_size), 1)
            bar = '=' * filled_len + '-' * (self.bar_length - filled_len)
            sys.stdout.write('[%s] %s%s|%s/%s ...\r' % (bar, percents, '%', self.current_size, self.total_size))
            sys.stdout.flush()
        else:
            filled_len = int(self.bar_length)
            bar = '=' * filled_len
            sys.stdout.write('[%s] %s%s|%s/%s ...\r' % (bar, 100.0, '%', self.total_size, self.total_size))
            sys.stdout.flush()
            sys.stdout.write("\n")


"""
Functions for detecting ECs that should not be
included in 3D structure prediction

Most functions in this module are rewritten from
older pipeline code in choose_CNS_constraint_set.m

Authors:
  Thomas A. Hopf
"""

from collections import defaultdict
from itertools import combinations, product

import pandas as pd
import numpy as np

from evcouplings.compare.pdb import ClassicPDB
from evcouplings.utils.calculations import dihedral_angle
from evcouplings.utils.config import InvalidParameterError
from evcouplings.visualize.pairs import find_secondary_structure_segments


def _alpha_dihedrals(coords, segments):
    """
    Compute dihedral score for alpha-helical
    segments.

    This function reimplements the functionality of
    analyze_alpha_helix.pl and make_alpha_beta_score_table.m
    (analyze_alpha_helix function) from the original pipeline.

    Parameters
    ----------
    coords : pandas.DataFrame
        Coordinate and residue information
        dataframe (residues with CA
        coordinates only)
    segments : list
        List of tuples (start, end) that
        define secondary structure segments
        (end index is exclusive)

    Returns
    -------
    pandas.DataFrame
        Table with dihedral angles for all
        helical residues with defined angle
        (columns: helix (helix index), i (position),
        dihedral (dihedral angle))
    """
    # check if we have cooords for position
    def has(pos):
        return pos in coords.i.values

    # return coords for position
    def xyz(pos):
        return coords.loc[
            coords.i == pos, ["x", "y", "z"]
        ].iloc[0].values

    res = []
    # go through all helical segments
    for helix_idx, (start, end) in enumerate(segments):
        # compute dihedral for each residue
        # within that helical segment
        for i in range(start, end):
            # check if all four necessary CA atoms are defined
            if has(i - 1) and has(i) and has(i + 1) and has(i + 2):
                # compute dihedral
                angle = dihedral_angle(
                    xyz(i - 1), xyz(i), xyz(i + 1), xyz(i + 2)
                )
                res.append((helix_idx, i, angle))

    return pd.DataFrame(
        res, columns=["helix", "i", "dihedral"]
    )


def _beta_dihedrals(coords, segments, max_strand_distance=7,
                    original=True):
    """
    Compute dihedral score for beta strand
    segments.

    This function reimplements the functionality of
    analyze_beta_strand.pl and make_alpha_beta_score_table.m
    (analyze_beta_strand function) from the original pipeline.

    Parameters
    ----------
    coords : pandas.DataFrame
        Coordinate and residue information
        dataframe (residues with CA
        coordinates only)
    segments : list
        List of tuples (start, end) that
        define secondary structure segments
        (end index is exclusive)
    max_strand_distance : float, optional (default: 7)
        Maximum distance of strands to be considered
        for dihedral angle calculation
    original : bool, optional (default: True):
        Use exact implementation of 2011 ranking protocol

    Returns
    -------
    pandas.DataFrame
        Table with dihedral angles for all
        sheet residues with defined angle
        (columns: helix (helix index), i (position),
        dihedral (dihedral angle))
    """
    # transfer coords into a dict of numpy vectors for faster access
    coords = dict(
        zip(coords.i, coords.loc[:, ["x", "y", "z"]].values)
    )

    # check if we have cooords for position
    def has(pos):
        return pos in coords

    # return coords for position
    def xyz(pos):
        return coords[pos]

    # determine orientation of strands (parallel or antiparallel)
    def _orientation(pairs):
        # votes in favor of parallel arrangement, and total votes
        par, total = 0, 0
        for idx, r in pairs.iterrows():
            i, j = r["pos_i"], r["pos_j"]
            if has(i - 2) and has(i + 2) and has(j + 2):
                d_par = np.linalg.norm(xyz(i + 2) - xyz(j + 2))

                if original:
                    # the residues chosen here disagree with residues
                    # for angle calculation
                    d_anti = np.linalg.norm(xyz(i - 2) - xyz(j + 2))
                else:
                    # alternative implementation consistent with angle
                    # calculation
                    d_anti = np.linalg.norm(xyz(i + 2) - xyz(j - 2))

                # cast votes
                total += 1
                if d_par <= d_anti:
                    par += 1

        if total == 0:
            return 0
        else:
            return par / total * 2 - 1

    # compute dihedral angles for residue pairings
    def _compute_dihedral(pairs, strands_parallel):
        res = []
        for idx, r in pairs.iterrows():
            i, j = r["pos_i"], r["pos_j"]
            if original:
                # the residues evaluated here disagree with residues
                # used for angle calculation
                if not has(j + 2):
                    continue

                if strands_parallel and not has(i + 2):
                    continue

                if not strands_parallel and not has(i - 2):
                    continue
            else:
                # alternative implementation consistent with angle
                # calculation
                if not has(i + 2):
                    continue

                if strands_parallel and not has(j + 2):
                    continue

                if not strands_parallel and not has(j - 2):
                    continue

            if strands_parallel:
                angle = dihedral_angle(
                    xyz(i), xyz(i + 2), xyz(j + 2), xyz(j)
                )
            else:
                angle = dihedral_angle(
                    xyz(i), xyz(i + 2), xyz(j - 2), xyz(j)
                )

            res.append(
                (i, j, int(r["strand_i"]), int(r["strand_j"]), angle)
            )

        return pd.DataFrame(
            res, columns=["i", "j", "strand_i", "strand_j", "dihedral"]
        )

    # go through all possible combinations of strands
    # and identify what possible partners are in 3D
    strand_partners = defaultdict(list)

    for (strand_i, seg_i), (strand_j, seg_j) in combinations(enumerate(segments), 2):
        # compute all pairwise C_alpha distances between residues in
        # strands i and j
        pair_dists = pd.DataFrame(
            [
                (
                    strand_i, strand_j, pos_i, pos_j,
                    np.linalg.norm(xyz(pos_i) - xyz(pos_j))
                )
                for (pos_i, pos_j) in product(range(*seg_i), range(*seg_j))
                if has(pos_i) and has(pos_j)
            ], columns=["strand_i", "strand_j", "pos_i", "pos_j", "dist"]
        )
        # test if strands are proximal, otherwise skip pair
        if pair_dists.dist.min() > max_strand_distance:
            continue

        # for each position in strand i,
        # identify nearest residue in strand j
        nearest = pair_dists.sort_values(
            by=["pos_i", "dist"]
        ).groupby(["pos_i"]).first().reset_index()

        # identify pairs that are close according to distance threshold
        close = nearest.loc[nearest.dist <= max_strand_distance]

        # also determine strand distance as closest of all pairs
        strand_dist = close.dist.min()

        # select anything between first and last close pair, this
        # defines the strand segments for further computations including angles
        first_good, last_good = close.index.min(), close.index.max()
        dihedral_pairs = nearest.loc[first_good:last_good]

        # determine if pair is parallel or antiparallel
        orientation_vote = _orientation(dihedral_pairs)
        strands_parallel = orientation_vote > 0

        # compute dihedrals (if we have the positions in coordinates)
        # print("ORIENTATION:", orientation_vote, strands_parallel)
        dihedrals = _compute_dihedral(dihedral_pairs, strands_parallel)

        # store strand pairing
        strand_partners[strand_i].append(
            (strand_dist, strand_j, dihedrals)
        )

    # now assemble final table of dihedrals,
    # allow only up to two pairings per strand
    # (if more, take the two closest ones in 3D based on strand_dist)
    all_dihedrals = pd.DataFrame()
    num_partners = defaultdict(int)
    for strand_i, partners in strand_partners.items():
        # only take two closest strands
        for dist, strand_j, dihedrals in sorted(partners):
            # check we do not have two partners yet
            if num_partners[strand_i] < 2 and num_partners[strand_j] < 2:
                all_dihedrals = all_dihedrals.append(dihedrals)
                num_partners[strand_i] += 1
                num_partners[strand_j] += 1

    return all_dihedrals


def dihedral_ranking_score(structure, residues, sec_struct_column="sec_struct_3state",
                           original=True):
    """
    Assess quality of structure model by twist of 
    predicted alpha-helices and beta-sheets.

    This function re-implements the functionality of
    make_alpha_beta_score_table.m from the original
    pipeline.

    Parameters
    ----------
    structure : evcouplings.compare.pdb.Chain
        Chain with 3D structure coordinates to evaluate
    residues : pandas.DataFrame
        Residue table with secondary structure predictions
        (columns i, A_i and secondary structure column)
    sec_struct_column : str, optional (default: sec_struct_3state)
        Column in residues dataframe that contains predicted
        secondary structure (H, E, C)
    original : bool, optional (default: True):
        Use exact implementation of 2011 ranking protocol

    Returns
    -------
    int
        Number of alpha dihedrals computed
    float
        Alpha dihedral score (unnormalized)
    int
        Number of beta dihedrals computed
    float
        Beta dihedral score (unnormalized)
    """
    # create table that, for each residue, contains
    # secondary structure and C_alpha coordinates.

    # First, throw away anything but C_alpha atoms
    structure = structure.filter_atoms(atom_name="CA")

    # Then, merge residue with atom information
    x = structure.residues.merge(
        structure.coords, left_index=True,
        right_on="residue_index"
    )

    # Then, merge with secondary structure prediction
    # PDB indices are strings, so merge on string
    residues = residues.copy()
    residues.loc[:, "id"] = residues.i.astype(str)
    x = residues.merge(
        x, on="id", how="left", suffixes=("", "_")
    )

    # find secondary structure segments
    _, _, segments = find_secondary_structure_segments(
        "".join(x.loc[:, sec_struct_column]), offset=x.i.min()
    )

    def _get_segments(seg_type):
        return [
            (start, end) for (type_, start, end) in segments
            if type_ == seg_type
        ]

    segs_alpha = _get_segments("H")
    segs_beta = _get_segments("E")

    # extract positions that actually have C_alpha coordinates
    x_valid = x.dropna(subset=["x", "y", "z"])

    # compute alpha helix and beta sheet dihedrals
    d_alpha = _alpha_dihedrals(x_valid, segs_alpha)
    d_beta = _beta_dihedrals(x_valid, segs_beta, original=original)

    # Finally, merge results into score

    # first, count how many angles in the right range
    # we have, but weight for actual value of the dihedral

    # alpha
    alpha_weights = [
        (0.2, 0.44, 0.52),
        (0.4, 0.52, 0.61),
        (0.6, 0.61, 0.70),
        (0.8, 0.70, 0.78),
        (1.0, 0.78, 0.96),
        (0.8, 0.96, 1.05),
        (0.6, 1.05, 1.13),
        (0.4, 1.13, 1.22),
        (0.2, 1.22, 1.31),
    ]

    if len(d_alpha) > 0:
        alpha_dihedral_score = sum([
            weight * len(d_alpha.query(
                "@lower < dihedral and dihedral <= @upper")
            )
            for weight, lower, upper in alpha_weights
        ])
    else:
        alpha_dihedral_score = 0

    # beta
    beta_weights = [
        (0.2, -0.3, -0.1),
        (0.4, -0.4, -0.3),
        (0.6, -0.5, -0.4),
        (0.8, -0.6, -0.5),
        (1.0, -0.8, -0.6),
        (0.8, -0.9, -0.8),
        (0.6, -1.0, -0.9),
        (0.4, -1.1, -1.0),
        (0.2, -1.2, -1.1),
    ]

    if len(d_beta) > 0:
        beta_dihedral_score = sum([
            weight * len(d_beta.query(
                "@lower <= dihedral and dihedral < @upper")
            )
            for weight, lower, upper in beta_weights
        ])
    else:
        beta_dihedral_score = 0

    return len(d_alpha), alpha_dihedral_score, len(d_beta), beta_dihedral_score


def dihedral_ranking(structure_files, residues, chain=None,
                     sec_struct_column="sec_struct_3state",
                     model=0):
    """
    Assess quality of a set of structure models by
    twist of predicted alpha-helices and beta-sheets.

    This function re-implements the final score table
    computed in make_alpha_beta_score_table.m from
    the original pipeline. Some of the implementation
    details where however modified, possibly leading
    to differences in the final computed ranking scores.

    Parameters
    ----------
    structure_files : list(str)
        Paths to PDB files that will be ranked
        (have to be in .pdb format)
    residues : pandas.DataFrame
        Residue table with secondary structure predictions
        (columns i, A_i and secondary structure column)
    chain : str, optional (default: None)
        Use this chain in each structure for the calculation.
        If None, will pick the only existing chain (if there
        are multiple chains, an InvalidParameterError will be
        raised).
    sec_struct_column : str, optional (default: sec_struct_3state)
        Column in residues dataframe that contains predicted
        secondary structure (H, E, C)
    model : int, optional (default: 0)
        Use this model from each PDB structure
        
    Returns
    -------
    pandas.DataFrame
        Table with final ranking scores (column ranking_score)
        and alpha and beta dihedral scores, as well
        as well as possible maximum score (equal to the number
        of computed dihedrals).
        
    Raises
    ------
    InvalidParameterError
        If chain is None but multiple chains exist in any
        of the structures
    """
    res = []
    for filename in structure_files:
        # load structure
        struc = ClassicPDB.from_file(filename)

        # see if we can select the right chain
        if chain is None:
            chains = struc.model_to_chains[model]
            if len(chains) != 1:
                raise InvalidParameterError(
                    "Model has more than one chain, need to "
                    "specify chain parameter to disambiguate."
                )
            chain = chains[0]

        # extract chain from structure
        sel_chain = struc.get_chain(chain)

        # compute dihedral ranking score
        x = dihedral_ranking_score(
            sel_chain, residues, sec_struct_column, original=False
        )
        res.append((filename, *x))

    r = pd.DataFrame(
        res, columns=[
            "filename", "num_alpha_dihedrals", "alpha_dihedral_score",
            "num_beta_dihedrals", "beta_dihedral_score"
        ]
    )

    # maximum score we could have obtained for either set of dihedrals
    max_alpha = r.num_alpha_dihedrals.max()

    # note that, unlike for alpha dihedrals, beta dihedrals are dependent
    # on what contacts between strands are made in 3D structure models
    max_beta = r.num_beta_dihedrals.max()

    # compute final ranking score
    # this computation is somewhat diffeent from original implementation:
    # - normalization for alpha helices is through number of possible dihedrals
    #   not the number of residues with helix secondary structure
    # - scores for helix and beta dihedrals are not adjusted to values < 0 in
    #   borderline cases

    # make sure we do not divide by 0 if we didn't count any dihedrals at all
    max_val = max(1, max_alpha + max_beta)

    r.loc[:, "ranking_score"] = (
        (r.alpha_dihedral_score + r.beta_dihedral_score) / max_val
    )

    return r
